#!/usr/bin/env python
"""
This file contains Python command example for the FLEX ACL module.
Python commands syntax is very similar to the Switch SDK APIs.
You can learn more about each command and its parameters by reading the SwitchX API Reference Manual.
"""
from __future__ import print_function
import sys
import socket
import struct
import errno
import os
from python_sdk_api.sx_api import *
from test_infra_common import *


INVALID_GROUP_ID = 0xFFFFFFFF
dir_dict = {
    SX_ACL_DIRECTION_INGRESS: 'INGRESS',
    SX_ACL_DIRECTION_EGRESS: 'EGRESS',
    SX_ACL_DIRECTION_RIF_INGRESS: 'RIF_INGRESS',
    SX_ACL_DIRECTION_RIF_EGRESS: 'RIF_EGRESS'
}

key_type_dict = {
    FLEX_ACL_KEY_INVALID: 'INVALID',
    FLEX_ACL_KEY_DMAC: 'DMAC',
    FLEX_ACL_KEY_SMAC: 'SMAC',
    FLEX_ACL_KEY_ETHERTYPE: 'ETHERTYPE',
    FLEX_ACL_KEY_DEI: 'DEI',
    FLEX_ACL_KEY_PCP: 'PCP',
    FLEX_ACL_KEY_VLAN_ID: 'VLAN_ID',
    FLEX_ACL_KEY_INNER_DMAC: 'INNER_DMAC',
    FLEX_ACL_KEY_INNER_SMAC: 'INNER_SMAC',
    FLEX_ACL_KEY_INNER_DEI: 'INNER_DEI',
    FLEX_ACL_KEY_INNER_PCP: 'INNER_PCP',
    FLEX_ACL_KEY_INNER_ETHERTYPE: 'INNER_ETHERTYPE',
    FLEX_ACL_KEY_DIP: 'DIP',
    FLEX_ACL_KEY_SIP: 'SIP',
    FLEX_ACL_KEY_DSCP: 'DSCP',
    FLEX_ACL_KEY_ECN: 'ECN',
    FLEX_ACL_KEY_IP_PACKET_LENGTH: 'IP_PACKET_LENGTH',
    FLEX_ACL_KEY_IP_PROTO: 'IP_PROTO',
    FLEX_ACL_KEY_TTL: 'TTL',
    FLEX_ACL_KEY_DIPV6: 'DIPV6',
    FLEX_ACL_KEY_SIPV6: 'SIPV6',
    FLEX_ACL_KEY_INNER_SIP: 'INNER_SIP',
    FLEX_ACL_KEY_INNER_DIP: 'INNER_DIP',
    FLEX_ACL_KEY_INNER_DSCP: 'INNER_DSCP',
    FLEX_ACL_KEY_INNER_ECN: 'INNER_ECN',
    FLEX_ACL_KEY_INNER_IP_PROTO: 'INNER_IP_PROTO',
    FLEX_ACL_KEY_INNER_SIPV6: 'INNER_SIPV6',
    FLEX_ACL_KEY_INNER_DIPV6: 'INNER_DIPV6',
    FLEX_ACL_KEY_L4_DESTINATION_PORT: 'L4_DESTINATION_PORT',
    FLEX_ACL_KEY_L4_SOURCE_PORT: 'L4_SOURCE_PORT',
    FLEX_ACL_KEY_TCP_CONTROL: 'TCP_CONTROL',
    FLEX_ACL_KEY_TCP_ECN: 'TCP_ECN',
    FLEX_ACL_KEY_INNER_L4_DESTINATION_PORT: 'INNER_L4_DESTINATION_PORT',
    FLEX_ACL_KEY_INNER_L4_SOURCE_PORT: 'INNER_L4_SOURCE_PORT',
    FLEX_ACL_KEY_GRE_KEY: 'GRE_KEY',
    FLEX_ACL_KEY_VNI_KEY: 'VNI_KEY',
    FLEX_ACL_KEY_GRE_PROTOCOL: 'GRE_PROTOCOL',
    FLEX_ACL_KEY_MPLS_LABEL_ID_1: 'MPLS_LABEL_ID_1',
    FLEX_ACL_KEY_MPLS_LABEL_ID_2: 'MPLS_LABEL_ID_2',
    FLEX_ACL_KEY_MPLS_LABEL_ID_3: 'MPLS_LABEL_ID_3',
    FLEX_ACL_KEY_MPLS_LABEL_ID_4: 'MPLS_LABEL_ID_4',
    FLEX_ACL_KEY_MPLS_LABEL_ID_5: 'MPLS_LABEL_ID_5',
    FLEX_ACL_KEY_MPLS_LABEL_ID_6: 'MPLS_LABEL_ID_6',
    FLEX_ACL_KEY_EXP: 'EXP',
    FLEX_ACL_KEY_BOS: 'BOS',
    FLEX_ACL_KEY_MPLS_TTL: 'MPLS_TTL',
    FLEX_ACL_KEY_RW_PCP: 'RW_PCP',
    FLEX_ACL_KEY_L2_DMAC_TYPE: 'L2_DMAC_TYPE',
    FLEX_ACL_KEY_DMAC_IS_UC: 'DMAC_IS_UC',
    FLEX_ACL_KEY_VLAN_TAGGED: 'VLAN_TAGGED',
    FLEX_ACL_KEY_VLAN_VALID: 'VLAN_VALID',
    FLEX_ACL_KEY_DST_PORT: 'DST_PORT',
    FLEX_ACL_KEY_SRC_PORT: 'SRC_PORT',
    FLEX_ACL_KEY_RW_DSCP: 'RW_DSCP',
    FLEX_ACL_KEY_IP_FRAGMENTED: 'IP_FRAGMENTED',
    FLEX_ACL_KEY_IP_DONT_FRAGMENT: 'IP_DONT_FRAGMENT',
    FLEX_ACL_KEY_IP_FRAGMENT_NOT_FIRST: 'IP_FRAGMENT_NOT_FIRST',
    FLEX_ACL_KEY_IP_OK: 'IP_OK',
    FLEX_ACL_KEY_IS_ARP: 'IS_ARP',
    FLEX_ACL_KEY_IP_OPT: 'IP_OPT',
    FLEX_ACL_KEY_IS_IP_V4: 'IS_IP_V4',
    FLEX_ACL_KEY_L3_TYPE: 'L3_TYPE',
    FLEX_ACL_KEY_TTL_OK: 'TTL_OK',
    FLEX_ACL_KEY_L4_OK: 'L4_OK',
    FLEX_ACL_KEY_L4_TYPE: 'L4_TYPE',
    FLEX_ACL_KEY_SWITCH_PRIO: 'SWITCH_PRIO',
    FLEX_ACL_KEY_COLOR: 'COLOR',
    FLEX_ACL_KEY_BUFF: 'BUFF',
    FLEX_ACL_KEY_L4_PORT_RANGE: 'L4_PORT_RANGE',
    FLEX_ACL_KEY_DISCARD_STATE: 'DISCARD_STATE',
    FLEX_ACL_KEY_IS_TRAPPED: 'IS_TRAPPED',
    FLEX_ACL_KEY_RX_LIST: 'RX_LIST',
    FLEX_ACL_KEY_IRIF: 'IRIF',
    FLEX_ACL_KEY_ERIF: 'ERIF',
    FLEX_ACL_KEY_VIRTUAL_ROUTER: 'VIRTUAL_ROUTER',
    FLEX_ACL_KEY_IPV6_EXTENSION_HEADERS: 'IPV6_EXTENSION_HEADERS',
    FLEX_ACL_KEY_IPV6_EXTENSION_HEADER_EXISTS: 'IPV6_EXTENSION_HEADER_EXISTS',
    FLEX_ACL_KEY_USER_TOKEN: 'USER_TOKEN',
    FLEX_ACL_KEY_INNER_VLAN_VALID: 'INNER_VLAN_VALID',
    FLEX_ACL_KEY_INNER_IP_OK: 'INNER_IP_OK',
    FLEX_ACL_KEY_INNER_L3_TYPE: 'INNER_L3_TYPE',
    FLEX_ACL_KEY_INNER_L4_OK: 'INNER_L4_OK',
    FLEX_ACL_KEY_INNER_TTL_OK: 'INNER_TTL_OK',
    FLEX_ACL_KEY_TUNNEL_TYPE: 'TUNNEL_TYPE',
    FLEX_ACL_KEY_GRE_KEY_EXISTS: 'GRE_KEY_EXISTS',
    FLEX_ACL_KEY_TUNNEL_NVE_TYPE: 'TUNNEL_NVE_TYPE',
    FLEX_ACL_KEY_L4_TYPE_EXTENDED: 'L4_TYPE_EXTENDED',
    FLEX_ACL_KEY_IS_MPLS: 'IS_MPLS',
    FLEX_ACL_KEY_MPLS_LABELS_VALID: 'MPLS_LABELS_VALID',
    FLEX_ACL_KEY_RW_EXP: 'RW_EXP',
    FLEX_ACL_KEY_RX_PORT_LIST: 'RX_PORT_LIST',
    FLEX_ACL_KEY_TX_PORT_LIST: 'TX_PORT_LIST',
    FLEX_ACL_KEY_IS_ROUTED: 'IS_ROUTED',
    FLEX_ACL_KEY_CUSTOM_BYTE_0: 'CUSTOM_BYTE_0',
    FLEX_ACL_KEY_CUSTOM_BYTE_1: 'CUSTOM_BYTE_1',
    FLEX_ACL_KEY_CUSTOM_BYTE_2: 'CUSTOM_BYTE_2',
    FLEX_ACL_KEY_CUSTOM_BYTE_3: 'CUSTOM_BYTE_3',
    FLEX_ACL_KEY_CUSTOM_BYTE_4: 'CUSTOM_BYTE_4',
    FLEX_ACL_KEY_CUSTOM_BYTE_5: 'CUSTOM_BYTE_5',
    FLEX_ACL_KEY_CUSTOM_BYTE_6: 'CUSTOM_BYTE_6',
    FLEX_ACL_KEY_CUSTOM_BYTE_7: 'CUSTOM_BYTE_7',
    FLEX_ACL_KEY_CUSTOM_BYTE_8: 'CUSTOM_BYTE_8',
    FLEX_ACL_KEY_CUSTOM_BYTE_9: 'CUSTOM_BYTE_9',
    FLEX_ACL_KEY_CUSTOM_BYTE_10: 'CUSTOM_BYTE_10',
    FLEX_ACL_KEY_CUSTOM_BYTE_11: 'CUSTOM_BYTE_11',
    FLEX_ACL_KEY_CUSTOM_BYTE_12: 'CUSTOM_BYTE_12',
    FLEX_ACL_KEY_CUSTOM_BYTE_13: 'CUSTOM_BYTE_13',
    FLEX_ACL_KEY_CUSTOM_BYTE_14: 'CUSTOM_BYTE_14',
    FLEX_ACL_KEY_CUSTOM_BYTE_15: 'CUSTOM_BYTE_15'
}

action_type_dict = {
    SX_FLEX_ACL_ACTION_FORWARD: 'FORWARD',
    SX_FLEX_ACL_ACTION_TRAP: 'TRAP',
    SX_FLEX_ACL_ACTION_COUNTER: 'COUNTER',
    SX_FLEX_ACL_ACTION_MIRROR: 'MIRROR',
    SX_FLEX_ACL_ACTION_POLICER: 'POLICER',
    SX_FLEX_ACL_ACTION_SET_PRIO: 'SET_PRIO',
    SX_FLEX_ACL_ACTION_SET_VLAN: 'SET_VLAN',
    SX_FLEX_ACL_ACTION_SET_INNER_VLAN_PRI: 'SET_INNER_VLAN_PRI',
    SX_FLEX_ACL_ACTION_SET_OUTER_VLAN_PRI: 'SET_OUTER_VLAN_PRI',
    SX_FLEX_ACL_ACTION_SET_INNER_VLAN_ID: 'SET_INNER_VLAN_ID',
    SX_FLEX_ACL_ACTION_SET_OUTER_VLAN_ID: 'SET_OUTER_VLAN_ID',
    SX_FLEX_ACL_ACTION_SET_SRC_MAC: 'SET_SRC_MAC',
    SX_FLEX_ACL_ACTION_SET_DST_MAC: 'SET_DST_MAC',
    SX_FLEX_ACL_ACTION_SET_DSCP: 'SET_DSCP',
    SX_FLEX_ACL_ACTION_SET_BRIDGE: 'SET_BRIDGE',
    SX_FLEX_ACL_ACTION_PBS: 'PBS',
    SX_FLEX_ACL_ACTION_SET_TC: 'SET_TC',
    SX_FLEX_ACL_ACTION_SET_TTL: 'SET_TTL',
    SX_FLEX_ACL_ACTION_DEC_TTL: 'DEC_TTL',
    SX_FLEX_ACL_ACTION_SET_COLOR: 'SET_COLOR',
    SX_FLEX_ACL_ACTION_SET_ECN: 'SET_ECN',
    SX_FLEX_ACL_ACTION_SET_USER_TOKEN: 'SET_USER_TOKEN',
    SX_FLEX_ACL_ACTION_DONT_LEARN: 'DONT_LEARN',
    SX_FLEX_ACL_ACTION_RPF: 'RPF',
    SX_FLEX_ACL_ACTION_MC_ROUTE: 'MC_ROUTE',
    SX_FLEX_ACL_ACTION_TUNNEL_DECAP: 'TUNNEL_DECAP',
    SX_FLEX_ACL_ACTION_GOTO: 'GOTO',
    SX_FLEX_ACL_ACTION_SET_ROUTER: 'SET_ROUTER',
    SX_FLEX_ACL_ACTION_MC: 'MC',
    SX_FLEX_ACL_ACTION_UC_ROUTE: 'UC_ROUTE',
    SX_FLEX_ACL_ACTION_SET_DSCP_REWRITE: 'SET_DSCP_REWRITE',
    SX_FLEX_ACL_ACTION_SET_PCP_REWRITE: 'SET_PCP_REWRITE',
    SX_FLEX_ACL_ACTION_EGRESS_MIRROR: 'EGRESS_MIRROR',
    SX_FLEX_ACL_ACTION_IGNORE_EGRESS_VLAN_FILTER: 'IGNORE_EGRESS_VLAN_FILTER',
    SX_FLEX_ACL_ACTION_IGNORE_EGRESS_STP_FILTER: 'IGNORE_EGRESS_STP_FILTER',
    SX_FLEX_ACL_ACTION_DISABLE_OVERLAY_LEARNING: 'DISABLE_OVERLAY_LEARNING',
    SX_FLEX_ACL_ACTION_PORT_FILTER: 'PORT_FILTER',
    SX_FLEX_ACL_ACTION_SET_MPLS_TTL: 'SET_MPLS_TTL',
    SX_FLEX_ACL_ACTION_DEC_MPLS_TTL: 'DEC_MPLS_TTL',
    SX_FLEX_ACL_ACTION_SET_EXP: 'SET_EXP',
    SX_FLEX_ACL_ACTION_SET_EXP_REWRITE: 'SET_EXP_REWRITE',
    SX_FLEX_ACL_ACTION_PBILM: 'PBILM',
    SX_FLEX_ACL_ACTION_NVE_TUNNEL_ENCAP: 'NVE_TUNNEL_ENCAP',
    SX_FLEX_ACL_ACTION_NVE_MC_TUNNEL_ENCAP: 'NVE_MC_TUNNEL_ENCAP',
    SX_FLEX_ACL_ACTION_TRAP_W_USER_ID: 'TRAP_W_USER_ID',
}

l2_dmac_type_dict = {
    SX_ACL_L2_DMAC_TYPE_MULTICAST: 'MULTICAST',
    SX_ACL_L2_DMAC_TYPE_BROADCAST: 'BROADCAST',
    SX_ACL_L2_DMAC_TYPE_UNICAST: 'UNICAST'
}

l3_type_dict = {
    SX_ACL_L3_TYPE_IPV4: 'IPV4',
    SX_ACL_L3_TYPE_IPV6: 'IPV6',
    SX_ACL_L3_TYPE_ARP: 'ARP',
    SX_ACL_L3_TYPE_OTHER: 'OTHER'
}

l4_type_dict = {
    SX_ACL_L4_TYPE_INVALID: 'INVALID',
    SX_ACL_L4_TYPE_TCP: 'TCP',
    SX_ACL_L4_TYPE_UDP: 'UDP',
    SX_ACL_L4_TYPE_OTHER: 'OTHER'
}

forward_action_dict = {
    SX_ACL_TRAP_FORWARD_ACTION_TYPE_FORWARD: 'FORWARD',
    SX_ACL_TRAP_FORWARD_ACTION_TYPE_DISCARD: 'DISCARD',
    SX_ACL_TRAP_FORWARD_ACTION_TYPE_SOFT_DISCARD: 'SOFT_DISCARD',
    SX_ACL_TRAP_FORWARD_ACTION_TYPE_PERMIT: 'PERMIT',
}

trap_action_dict = {
    SX_ACL_TRAP_ACTION_TYPE_TRAP: 'TRAP',
    SX_ACL_TRAP_ACTION_TYPE_DISCARD: 'DISCARD',
    SX_ACL_TRAP_ACTION_TYPE_SOFT_DISCARD: 'SOFT_DISCARD',
}

tunnel_type_dict = {
    SX_TUNNEL_TYPE_IPINIP_P2P_IPV4_IN_IPV4: 'IPINIP_P2P_IPV4_IN_IPV4',
    SX_TUNNEL_TYPE_IPINIP_P2P_IPV4_IN_GRE: 'IPINIP_P2P_IPV4_IN_GRE',

    SX_TUNNEL_TYPE_NVE_VXLAN: 'NVE_VXLAN',
    SX_TUNNEL_TYPE_NVE_VXLAN_GPE: 'NVE_VXLAN_GPE',
    SX_TUNNEL_TYPE_NVE_GENEVE: 'NVE_GENEVE',
    SX_TUNNEL_TYPE_NVE_NVGRE: 'NVE_NVGRE',
    SX_TUNNEL_TYPE_NVE_VXLAN_IPV6: 'NVE_VXLAN_IPV6',
}

l4_type_extended_dict = {
    SX_ACL_L4_TYPE_EXTENDED_TCP: 'TCP',
    SX_ACL_L4_TYPE_EXTENDED_UDP: 'UDP',
    SX_ACL_L4_TYPE_EXTENDED_BTH: 'BTH',
    SX_ACL_L4_TYPE_EXTENDED_BTHOUDP: 'BTHOUDP',
    SX_ACL_L4_TYPE_EXTENDED_ICMP: 'ICMP',
    SX_ACL_L4_TYPE_EXTENDED_IGMP: 'IGMP',
    SX_ACL_L4_TYPE_EXTENDED_AH: 'AH',
    SX_ACL_L4_TYPE_EXTENDED_ESP: 'ESP',
    SX_ACL_L4_TYPE_EXTENDED_OTHERS: 'OTHERS'
}

port_list_match_dict = {
    SX_ACL_PORT_LIST_MATCH_NEGATIVE: 'NEGATIVE',
    SX_ACL_PORT_LIST_MATCH_POSITIVE: 'POSITIVE'
}

action_set_vlan_type_dict = {
    SX_ACL_FLEX_SET_VLAN_CMD_TYPE_PUSH: 'PUSH',
    SX_ACL_FLEX_SET_VLAN_CMD_TYPE_POP: 'POP'
}

qinq_tunnel_qos_dict = {
    SX_ACL_FLEX_QINQ_TUNNEL_QOS_UNIFORM: 'UNIFORM',
    SX_ACL_FLEX_QINQ_TUNNEL_QOS_PIPE: 'PIPE'
}

action_color_type_dict = {
    SX_ACL_FLEX_COLOR_GREEN: 'GREEN',
    SX_ACL_FLEX_COLOR_YELLOW: 'YELLOW',
    SX_ACL_FLEX_COLOR_RED: 'RED'
}

rpf_action_dict = {
    SX_ACL_RPF_ACTION_TYPE_DISABLED: 'DISABLED',
    SX_ACL_RPF_ACTION_TYPE_DROP: 'DROP',
    SX_ACL_RPF_ACTION_TYPE_TRAP: 'TRAP'
}

action_goto_cmd_dict = {
    SX_ACL_ACTION_GOTO_JUMP: 'JUMP',
    SX_ACL_ACTION_GOTO_CALL: 'CALL',
    SX_ACL_ACTION_GOTO_BREAK: 'BREAK',
    SX_ACL_ACTION_GOTO_TERMINATE: 'TERMINATE'
}

action_rewrite_cmd_dict = {
    SX_ACL_ACTION_REWRITE_ENABLE: 'ENABLE',
    SX_ACL_ACTION_REWRITE_DISABLE: 'DISABLE'
}


def get_key_val_mask_str(key_desc):
    key_value = ""
    key_mask = ""

    if key_desc.key_id == FLEX_ACL_KEY_INVALID:
        key_value = "INVALID"
        key_mask = "INVALID"

    elif key_desc.key_id == FLEX_ACL_KEY_DMAC:
        key_value = key_desc.key.dmac.to_str()
        key_mask = key_desc.mask.dmac.to_str()

    elif key_desc.key_id == FLEX_ACL_KEY_SMAC:
        key_value = key_desc.key.smac.to_str()
        key_mask = key_desc.mask.smac.to_str()

    elif key_desc.key_id == FLEX_ACL_KEY_ETHERTYPE:
        key_value = "0x%x" % key_desc.key.ethertype
        key_mask = "0x%x" % key_desc.mask.ethertype

    elif key_desc.key_id == FLEX_ACL_KEY_DEI:
        key_value = "0x%x" % key_desc.key.dei
        key_mask = "0x%x" % key_desc.mask.dei

    elif key_desc.key_id == FLEX_ACL_KEY_PCP:
        key_value = "0x%x" % key_desc.key.pcp
        key_mask = "0x%x" % key_desc.mask.pcp

    elif key_desc.key_id == FLEX_ACL_KEY_VLAN_ID:
        key_value = "0x%x" % key_desc.key.vlan_id
        key_mask = "0x%x" % key_desc.mask.vlan_id

    elif key_desc.key_id == FLEX_ACL_KEY_INNER_DMAC:
        key_value = key_desc.key.inner_dmac.to_str()
        key_mask = key_desc.mask.inner_dmac.to_str()

    elif key_desc.key_id == FLEX_ACL_KEY_INNER_SMAC:
        key_value = key_desc.key.inner_smac.to_str()
        key_mask = key_desc.mask.inner_smac.to_str()

    elif key_desc.key_id == FLEX_ACL_KEY_INNER_DEI:
        key_value = "0x%x" % key_desc.key.inner_dei
        key_mask = "0x%x" % key_desc.mask.inner_dei

    elif key_desc.key_id == FLEX_ACL_KEY_INNER_PCP:
        key_value = "0x%x" % key_desc.key.inner_pcp
        key_mask = "0x%x" % key_desc.mask.inner_pcp

    elif key_desc.key_id == FLEX_ACL_KEY_INNER_ETHERTYPE:
        key_value = "0x%x" % key_desc.key.inner_ethertype
        key_mask = "0x%x" % key_desc.mask.inner_ethertype

    elif key_desc.key_id == FLEX_ACL_KEY_DIP:
        key_value = ip_addr_to_str(key_desc.key.dip)
        key_mask = ip_addr_to_str(key_desc.mask.dip)

    elif key_desc.key_id == FLEX_ACL_KEY_SIP:
        key_value = ip_addr_to_str(key_desc.key.sip)
        key_mask = ip_addr_to_str(key_desc.mask.sip)

    elif key_desc.key_id == FLEX_ACL_KEY_DSCP:
        key_value = "0x%x" % key_desc.key.dscp
        key_mask = "0x%x" % key_desc.mask.dscp

    elif key_desc.key_id == FLEX_ACL_KEY_ECN:
        key_value = "0x%x" % key_desc.key.ecn
        key_mask = "0x%x" % key_desc.mask.ecn

    elif key_desc.key_id == FLEX_ACL_KEY_IP_PACKET_LENGTH:
        key_value = "0x%x" % key_desc.key.ip_packet_length
        key_mask = "0x%x" % key_desc.mask.ip_packet_length

    elif key_desc.key_id == FLEX_ACL_KEY_IP_PROTO:
        key_value = "0x%x" % key_desc.key.ip_proto
        key_mask = "0x%x" % key_desc.mask.ip_proto

    elif key_desc.key_id == FLEX_ACL_KEY_TTL:
        key_value = "0x%x" % key_desc.key.ttl
        key_mask = "0x%x" % key_desc.mask.ttl

    elif key_desc.key_id == FLEX_ACL_KEY_DIPV6:
        key_value = ip_addr_to_str(key_desc.key.dipv6)
        key_mask = ip_addr_to_str(key_desc.mask.dipv6)

    elif key_desc.key_id == FLEX_ACL_KEY_SIPV6:
        key_value = ip_addr_to_str(key_desc.key.sipv6)
        key_mask = ip_addr_to_str(key_desc.mask.sipv6)

    elif key_desc.key_id == FLEX_ACL_KEY_INNER_SIP:
        key_value = ip_addr_to_str(key_desc.key.inner_sip)
        key_mask = ip_addr_to_str(key_desc.mask.inner_sip)

    elif key_desc.key_id == FLEX_ACL_KEY_INNER_DIP:
        key_value = ip_addr_to_str(key_desc.key.inner_dip)
        key_mask = ip_addr_to_str(key_desc.mask.inner_dip)

    elif key_desc.key_id == FLEX_ACL_KEY_INNER_DSCP:
        key_value = "0x%x" % key_desc.key.inner_dscp
        key_mask = "0x%x" % key_desc.mask.inner_dscp

    elif key_desc.key_id == FLEX_ACL_KEY_INNER_ECN:
        key_value = "0x%x" % key_desc.key.inner_ecn
        key_mask = "0x%x" % key_desc.mask.inner_ecn

    elif key_desc.key_id == FLEX_ACL_KEY_INNER_IP_PROTO:
        key_value = "0x%x" % key_desc.key.inner_ip_proto
        key_mask = "0x%x" % key_desc.mask.inner_ip_proto

    elif key_desc.key_id == FLEX_ACL_KEY_INNER_SIPV6:
        key_value = ip_addr_to_str(key_desc.key.inner_sipv6)
        key_mask = ip_addr_to_str(key_desc.mask.inner_sipv6)

    elif key_desc.key_id == FLEX_ACL_KEY_INNER_DIPV6:
        key_value = ip_addr_to_str(key_desc.key.inner_dipv6)
        key_mask = ip_addr_to_str(key_desc.mask.inner_dipv6)

    elif key_desc.key_id == FLEX_ACL_KEY_L4_DESTINATION_PORT:
        key_value = "0x%x" % key_desc.key.l4_destination_port
        key_mask = "0x%x" % key_desc.mask.l4_destination_port

    elif key_desc.key_id == FLEX_ACL_KEY_L4_SOURCE_PORT:
        key_value = "0x%x" % key_desc.key.l4_source_port
        key_mask = "0x%x" % key_desc.mask.l4_source_port

    elif key_desc.key_id == FLEX_ACL_KEY_TCP_CONTROL:
        key_value = "0x%x" % key_desc.key.tcp_control
        key_mask = "0x%x" % key_desc.mask.tcp_control

    elif key_desc.key_id == FLEX_ACL_KEY_TCP_ECN:
        key_value = "0x%x" % key_desc.key.tcp_ecn
        key_mask = "0x%x" % key_desc.mask.tcp_ecn

    elif key_desc.key_id == FLEX_ACL_KEY_INNER_L4_DESTINATION_PORT:
        key_value = "0x%x" % key_desc.key.inner_l4_destination_port
        key_mask = "0x%x" % key_desc.mask.inner_l4_destination_port

    elif key_desc.key_id == FLEX_ACL_KEY_INNER_L4_SOURCE_PORT:
        key_value = "0x%x" % key_desc.key.inner_l4_source_port
        key_mask = "0x%x" % key_desc.mask.inner_l4_source_port

    elif key_desc.key_id == FLEX_ACL_KEY_GRE_KEY:
        key_value = "0x%x" % key_desc.key.gre_key
        key_mask = "0x%x" % key_desc.mask.gre_key

    elif key_desc.key_id == FLEX_ACL_KEY_VNI_KEY:
        key_value = "0x%x" % key_desc.key.vni_key
        key_mask = "0x%x" % key_desc.mask.vni_key

    elif key_desc.key_id == FLEX_ACL_KEY_GRE_PROTOCOL:
        key_value = "0x%x" % key_desc.key.gre_protocol
        key_mask = "0x%x" % key_desc.mask.gre_protocol

    elif key_desc.key_id == FLEX_ACL_KEY_MPLS_LABEL_ID_1:
        key_value = "0x%x" % key_desc.key.mpls_label_id_1
        key_mask = "0x%x" % key_desc.mask.mpls_label_id_1

    elif key_desc.key_id == FLEX_ACL_KEY_MPLS_LABEL_ID_2:
        key_value = "0x%x" % key_desc.key.mpls_label_id_2
        key_mask = "0x%x" % key_desc.mask.mpls_label_id_2

    elif key_desc.key_id == FLEX_ACL_KEY_MPLS_LABEL_ID_3:
        key_value = "0x%x" % key_desc.key.mpls_label_id_3
        key_mask = "0x%x" % key_desc.mask.mpls_label_id_3

    elif key_desc.key_id == FLEX_ACL_KEY_MPLS_LABEL_ID_4:
        key_value = "0x%x" % key_desc.key.mpls_label_id_4
        key_mask = "0x%x" % key_desc.mask.mpls_label_id_4

    elif key_desc.key_id == FLEX_ACL_KEY_MPLS_LABEL_ID_5:
        key_value = "0x%x" % key_desc.key.mpls_label_id_5
        key_mask = "0x%x" % key_desc.mask.mpls_label_id_5

    elif key_desc.key_id == FLEX_ACL_KEY_MPLS_LABEL_ID_6:
        key_value = "0x%x" % key_desc.key.mpls_label_id_6
        key_mask = "0x%x" % key_desc.mask.mpls_label_id_6

    elif key_desc.key_id == FLEX_ACL_KEY_EXP:
        key_value = "0x%x" % key_desc.key.exp
        key_mask = "0x%x" % key_desc.mask.exp

    elif key_desc.key_id == FLEX_ACL_KEY_BOS:
        key_value = "0x%x" % key_desc.key.bos
        key_mask = "0x%x" % key_desc.mask.bos

    elif key_desc.key_id == FLEX_ACL_KEY_MPLS_TTL:
        key_value = "0x%x" % key_desc.key.mpls_ttl
        key_mask = "0x%x" % key_desc.mask.mpls_ttl

    elif key_desc.key_id == FLEX_ACL_KEY_RW_PCP:
        key_value = "0x%x" % key_desc.key.rw_pcp
        key_mask = "0x%x" % key_desc.mask.rw_pcp

    elif key_desc.key_id == FLEX_ACL_KEY_L2_DMAC_TYPE:
        if key_desc.key.l2_dmac_type in l2_dmac_type_dict:
            key_value = l2_dmac_type_dict[key_desc.key.l2_dmac_type]
            key_mask = "0x%x" % key_desc.mask.l2_dmac_type

    elif key_desc.key_id == FLEX_ACL_KEY_DMAC_IS_UC:
        key_value = "0x%x" % key_desc.key.dmac_is_uc
        key_mask = "0x%x" % key_desc.mask.dmac_is_uc

    elif key_desc.key_id == FLEX_ACL_KEY_VLAN_TAGGED:
        key_value = "0x%x" % key_desc.key.vlan_tagged
        key_mask = "0x%x" % key_desc.mask.vlan_tagged

    elif key_desc.key_id == FLEX_ACL_KEY_VLAN_VALID:
        key_value = "0x%x" % key_desc.key.vlan_valid
        key_mask = "0x%x" % key_desc.mask.vlan_valid

    elif key_desc.key_id == FLEX_ACL_KEY_DST_PORT:
        key_value = "0x%x" % key_desc.key.dst_port
        key_mask = "0x%x" % key_desc.mask.dst_port

    elif key_desc.key_id == FLEX_ACL_KEY_SRC_PORT:
        key_value = "0x%x" % key_desc.key.src_port
        key_mask = "0x%x" % key_desc.mask.src_port

    elif key_desc.key_id == FLEX_ACL_KEY_RW_DSCP:
        key_value = "0x%x" % key_desc.key.rw_dscp
        key_mask = "0x%x" % key_desc.mask.rw_dscp

    elif key_desc.key_id == FLEX_ACL_KEY_IP_FRAGMENTED:
        key_value = "0x%x" % key_desc.key.ip_fragmented
        key_mask = "0x%x" % key_desc.mask.ip_fragmented

    elif key_desc.key_id == FLEX_ACL_KEY_IP_DONT_FRAGMENT:
        key_value = "0x%x" % key_desc.key.ip_dont_fragment
        key_mask = "0x%x" % key_desc.mask.ip_dont_fragment

    elif key_desc.key_id == FLEX_ACL_KEY_IP_FRAGMENT_NOT_FIRST:
        key_value = "0x%x" % key_desc.key.ip_fragment_not_first
        key_mask = "0x%x" % key_desc.mask.ip_fragment_not_first

    elif key_desc.key_id == FLEX_ACL_KEY_IP_OK:
        key_value = "0x%x" % key_desc.key.ip_ok
        key_mask = "0x%x" % key_desc.mask.ip_ok

    elif key_desc.key_id == FLEX_ACL_KEY_IS_ARP:
        key_value = "0x%x" % key_desc.key.is_arp
        key_mask = "0x%x" % key_desc.mask.is_arp

    elif key_desc.key_id == FLEX_ACL_KEY_IP_OPT:
        key_value = "0x%x" % key_desc.key.ip_opt
        key_mask = "0x%x" % key_desc.mask.ip_opt

    elif key_desc.key_id == FLEX_ACL_KEY_IS_IP_V4:
        key_value = "0x%x" % key_desc.key.is_ip_v4
        key_mask = "0x%x" % key_desc.mask.is_ip_v4

    elif key_desc.key_id == FLEX_ACL_KEY_L3_TYPE:
        if key_desc.key.l3_type in l3_type_dict:
            key_value = l3_type_dict[key_desc.key.l3_type]
            key_mask = "0x%x" % (key_desc.mask.l3_type)

    elif key_desc.key_id == FLEX_ACL_KEY_TTL_OK:
        key_value = "0x%x" % key_desc.key.ttl_ok
        key_mask = "0x%x" % key_desc.mask.ttl_ok

    elif key_desc.key_id == FLEX_ACL_KEY_L4_OK:
        key_value = "0x%x" % key_desc.key.l4_ok
        key_mask = "0x%x" % key_desc.mask.l4_ok

    elif key_desc.key_id == FLEX_ACL_KEY_L4_TYPE:
        if key_desc.key.l4_type in l4_type_dict:
            key_value = l4_type_dict[key_desc.key.l4_type]
            key_mask = "0x%x" % (key_desc.mask.l4_type)

    elif key_desc.key_id == FLEX_ACL_KEY_SWITCH_PRIO:
        key_value = "0x%x" % key_desc.key.switch_prio
        key_mask = "0x%x" % key_desc.mask.switch_prio

    elif key_desc.key_id == FLEX_ACL_KEY_COLOR:
        key_value = "0x%x" % key_desc.key.color
        key_mask = "0x%x" % key_desc.mask.color

    elif key_desc.key_id == FLEX_ACL_KEY_BUFF:
        key_value = "0x%x" % key_desc.key.buff
        key_mask = "0x%x" % key_desc.mask.buff

    elif key_desc.key_id == FLEX_ACL_KEY_L4_PORT_RANGE:
        key_value += "0x"
        for i in xrange(0, RM_API_ACL_PORT_RANGES_MAX):
            key_value += "%x" % sx_acl_port_range_id_t_arr_getitem(key_desc.key.l4_port_range.port_range_list, i)
        key_mask = "0x%x" % key_desc.mask.l4_port_range

    elif key_desc.key_id == FLEX_ACL_KEY_DISCARD_STATE:
        if key_desc.key.discard_state in forward_action_dict:
            key_value = forward_action_dict[key_desc.key.discard_state]
            key_mask = "0x%x" % key_desc.mask.discard_state

    elif key_desc.key_id == FLEX_ACL_KEY_IS_TRAPPED:
        key_value = "0x%x" % key_desc.key.is_trapped
        key_mask = "0x%x" % key_desc.mask.is_trapped

    elif key_desc.key_id == FLEX_ACL_KEY_RX_LIST:
        key_value = "0x%x" % key_desc.key.rx_list
        key_mask = "0x%x" % key_desc.mask.rx_list

    elif key_desc.key_id == FLEX_ACL_KEY_IRIF:
        key_value = "0x%x" % key_desc.key.irif
        key_mask = "0x%x" % key_desc.mask.irif

    elif key_desc.key_id == FLEX_ACL_KEY_ERIF:
        key_value = "0x%x" % key_desc.key.erif
        key_mask = "0x%x" % key_desc.mask.erif

    elif key_desc.key_id == FLEX_ACL_KEY_VIRTUAL_ROUTER:
        key_value = "0x%x" % key_desc.key.virtual_router
        key_mask = "0x%x" % key_desc.mask.virtual_router

    elif key_desc.key_id == FLEX_ACL_KEY_IPV6_EXTENSION_HEADERS:
        key_value = "0x"
        for i in xrange(0, SX_FLEX_ACL_IPV6_EXTENSION_HEADER_LAST):
            key_value += "%x" % sx_flex_acl_ipv6_extension_header_t_arr_getitem(key_desc.key.ipv6_extension_headers.extension_headers_list, i)
        key_mask = "0x%x" % key_desc.mask.ipv6_extension_headers

    elif key_desc.key_id == FLEX_ACL_KEY_IPV6_EXTENSION_HEADER_EXISTS:
        key_value = "0x%x" % key_desc.key.ipv6_extension_header_exists
        key_mask = "0x%x" % key_desc.mask.ipv6_extension_header_exists

    elif key_desc.key_id == FLEX_ACL_KEY_USER_TOKEN:
        key_value = "0x%x" % key_desc.key.user_token
        key_mask = "0x%x" % key_desc.mask.user_token

    elif key_desc.key_id == FLEX_ACL_KEY_INNER_VLAN_VALID:
        key_value = "0x%x" % key_desc.key.inner_vlan_valid
        key_mask = "0x%x" % key_desc.mask.inner_vlan_valid

    elif key_desc.key_id == FLEX_ACL_KEY_INNER_IP_OK:
        key_value = "0x%x" % key_desc.key.inner_ip_ok
        key_mask = "0x%x" % key_desc.mask.inner_ip_ok

    elif key_desc.key_id == FLEX_ACL_KEY_INNER_L3_TYPE:
        if key_desc.key.inner_l3_type in l3_type_dict:
            key_value = l3_type_dict[key_desc.key.inner_l3_type]
            key_mask = "0x%x" % (key_desc.mask.inner_l3_type)

    elif key_desc.key_id == FLEX_ACL_KEY_INNER_L4_OK:
        key_value = "0x%x" % key_desc.key.inner_l4_ok
        key_mask = "0x%x" % key_desc.mask.inner_l4_ok

    elif key_desc.key_id == FLEX_ACL_KEY_INNER_TTL_OK:
        key_value = "0x%x" % key_desc.key.inner_ttl_ok
        key_mask = "0x%x" % key_desc.mask.inner_ttl_ok

    elif key_desc.key_id == FLEX_ACL_KEY_TUNNEL_TYPE:
        if key_desc.key.tunnel_type in tunnel_type_dict:
            key_value = tunnel_type_dict[key_desc.key.tunnel_type]
            key_mask = "0x%x" % key_desc.mask.tunnel_type

    elif key_desc.key_id == FLEX_ACL_KEY_GRE_KEY_EXISTS:
        key_value = "0x%x" % key_desc.key.gre_key_exists
        key_mask = "0x%x" % key_desc.mask.gre_key_exists

    elif key_desc.key_id == FLEX_ACL_KEY_TUNNEL_NVE_TYPE:
        if key_desc.key.tunnel_nve_type in tunnel_type_dict:
            key_value = tunnel_type_dict[key_desc.key.tunnel_nve_type]
            key_mask = "0x%x" % key_desc.mask.tunnel_nve_type

    elif key_desc.key_id == FLEX_ACL_KEY_L4_TYPE_EXTENDED:
        if key_desc.key.l4_type_extended in l4_type_extended_dict:
            key_value = l4_type_extended_dict[key_desc.key.l4_type_extended]
            key_mask = "0x%x" % (key_desc.mask.l4_type_extended)

    elif key_desc.key_id == FLEX_ACL_KEY_IS_MPLS:
        key_value = "0x%x" % key_desc.key.is_mpls
        key_mask = "0x%x" % key_desc.mask.is_mpls

    elif key_desc.key_id == FLEX_ACL_KEY_MPLS_LABELS_VALID:
        key_value = "0x%x" % key_desc.key.mpls_labels_valid
        key_mask = "0x%x" % key_desc.mask.mpls_labels_valid

    elif key_desc.key_id == FLEX_ACL_KEY_RW_EXP:
        key_value = "0x%x" % key_desc.key.rw_exp
        key_mask = "0x%x" % key_desc.mask.rw_exp

    elif key_desc.key_id == FLEX_ACL_KEY_RX_PORT_LIST:
        if key_desc.key.rx_port_list.match_type in port_list_match_dict:
            key_value = "Match Type: %s, MC Container ID: 0x%x" % (port_list_match_dict[key_desc.key.rx_port_list.match_type], key_desc.key.rx_port_list.mc_container_id)
            key_mask = "0x%x" % key_desc.mask.rx_port_list

    elif key_desc.key_id == FLEX_ACL_KEY_TX_PORT_LIST:
        if key_desc.key.tx_port_list.match_type in port_list_match_dict:
            key_value = "Match Type: %s, MC Container ID: 0x%x" % (port_list_match_dict[key_desc.key.tx_port_list.match_type], key_desc.key.tx_port_list.mc_container_id)
            key_mask = "0x%x" % key_desc.mask.tx_port_list

    elif key_desc.key_id == FLEX_ACL_KEY_IS_ROUTED:
        key_value = "0x%x" % key_desc.key.is_routed
        key_mask = "0x%x" % key_desc.mask.is_routed

    elif key_desc.key_id == FLEX_ACL_KEY_CUSTOM_BYTE_0:
        key_value = "0x%x" % key_desc.key.custom_byte
        key_mask = "0x%x" % key_desc.mask.custom_byte

    elif key_desc.key_id == FLEX_ACL_KEY_CUSTOM_BYTE_1:
        key_value = "0x%x" % key_desc.key.custom_byte
        key_mask = "0x%x" % key_desc.mask.custom_byte

    elif key_desc.key_id == FLEX_ACL_KEY_CUSTOM_BYTE_2:
        key_value = "0x%x" % key_desc.key.custom_byte
        key_mask = "0x%x" % key_desc.mask.custom_byte

    elif key_desc.key_id == FLEX_ACL_KEY_CUSTOM_BYTE_3:
        key_value = "0x%x" % key_desc.key.custom_byte
        key_mask = "0x%x" % key_desc.mask.custom_byte

    elif key_desc.key_id == FLEX_ACL_KEY_CUSTOM_BYTE_4:
        key_value = "0x%x" % key_desc.key.custom_byte
        key_mask = "0x%x" % key_desc.mask.custom_byte

    elif key_desc.key_id == FLEX_ACL_KEY_CUSTOM_BYTE_5:
        key_value = "0x%x" % key_desc.key.custom_byte
        key_mask = "0x%x" % key_desc.mask.custom_byte

    elif key_desc.key_id == FLEX_ACL_KEY_CUSTOM_BYTE_6:
        key_value = "0x%x" % key_desc.key.custom_byte
        key_mask = "0x%x" % key_desc.mask.custom_byte

    elif key_desc.key_id == FLEX_ACL_KEY_CUSTOM_BYTE_7:
        key_value = "0x%x" % key_desc.key.custom_byte
        key_mask = "0x%x" % key_desc.mask.custom_byte

    elif key_desc.key_id == FLEX_ACL_KEY_CUSTOM_BYTE_8:
        key_value = "0x%x" % key_desc.key.custom_byte
        key_mask = "0x%x" % key_desc.mask.custom_byte

    elif key_desc.key_id == FLEX_ACL_KEY_CUSTOM_BYTE_9:
        key_value = "0x%x" % key_desc.key.custom_byte
        key_mask = "0x%x" % key_desc.mask.custom_byte

    elif key_desc.key_id == FLEX_ACL_KEY_CUSTOM_BYTE_10:
        key_value = "0x%x" % key_desc.key.custom_byte
        key_mask = "0x%x" % key_desc.mask.custom_byte

    elif key_desc.key_id == FLEX_ACL_KEY_CUSTOM_BYTE_11:
        key_value = "0x%x" % key_desc.key.custom_byte
        key_mask = "0x%x" % key_desc.mask.custom_byte

    elif key_desc.key_id == FLEX_ACL_KEY_CUSTOM_BYTE_12:
        key_value = "0x%x" % key_desc.key.custom_byte
        key_mask = "0x%x" % key_desc.mask.custom_byte

    elif key_desc.key_id == FLEX_ACL_KEY_CUSTOM_BYTE_13:
        key_value = "0x%x" % key_desc.key.custom_byte
        key_mask = "0x%x" % key_desc.mask.custom_byte

    elif key_desc.key_id == FLEX_ACL_KEY_CUSTOM_BYTE_14:
        key_value = "0x%x" % key_desc.key.custom_byte
        key_mask = "0x%x" % key_desc.mask.custom_byte

    elif key_desc.key_id == FLEX_ACL_KEY_CUSTOM_BYTE_15:
        key_value = "0x%x" % key_desc.key.custom_byte
        key_mask = "0x%x" % key_desc.mask.custom_byte

    return (key_value, key_mask)


def get_action_fields_str(action):
    action_fields = ""

    if action.type == SX_FLEX_ACL_ACTION_FORWARD:
        if action.fields.action_forward.action in forward_action_dict:
            action_fields = "Action: %s" % forward_action_dict[action.fields.action_forward.action]

    elif action.type == SX_FLEX_ACL_ACTION_TRAP:
        if action.fields.action_trap.action in trap_action_dict:
            action_fields = "Action: %s, Trap ID: 0x%x" % (trap_action_dict[action.fields.action_trap.action], action.fields.action_trap.trap_id)

    elif action.type == SX_FLEX_ACL_ACTION_COUNTER:
        action_fields = "Counter ID: 0x%x" % action.fields.action_counter.counter_id

    elif action.type == SX_FLEX_ACL_ACTION_MIRROR:
        action_fields = "Session ID: 0x%x" % action.fields.action_mirror.session_id

    elif action.type == SX_FLEX_ACL_ACTION_POLICER:
        action_fields = "Policer ID: 0x%x" % action.fields.action_policer.policer_id

    elif action.type == SX_FLEX_ACL_ACTION_SET_PRIO:
        action_fields = "Prio: 0x%x" % action.fields.action_set_prio.prio_val

    elif action.type == SX_FLEX_ACL_ACTION_SET_VLAN:
        if action.fields.action_set_vlan.cmd in action_set_vlan_type_dict and action.fields.action_set_vlan.qinq_tunnel_qos in qinq_tunnel_qos_dict:
            action_fields = "CMD: %s, VLAN: 0x%x, QOS: %s, PCP: 0x%x, DEI: 0x%x" % (action_set_vlan_type_dict[action.fields.action_set_vlan.cmd],
                                                                                    action.fields.action_set_vlan.vlan_id,
                                                                                    qinq_tunnel_qos_dict[action.fields.action_set_vlan.qinq_tunnel_qos],
                                                                                    action.fields.action_set_vlan.qinq_tunnel_qos_fields.pipe.pcp,
                                                                                    action.fields.action_set_vlan.qinq_tunnel_qos_fields.pipe.dei)

    elif action.type == SX_FLEX_ACL_ACTION_SET_INNER_VLAN_PRI:
        action_fields = "PCP: 0x%x, DEI: 0x%x" % (action.fields.action_set_inner_vlan_prio.pcp, action.fields.action_set_inner_vlan_prio.dei)

    elif action.type == SX_FLEX_ACL_ACTION_SET_OUTER_VLAN_PRI:
        action_fields = "PCP: 0x%x, DEI: 0x%x" % (action.fields.action_set_outer_vlan_prio.pcp, action.fields.action_set_outer_vlan_prio.dei)

    elif action.type == SX_FLEX_ACL_ACTION_SET_INNER_VLAN_ID:
        action_fields = "VLAN: 0x%x" % action.fields.action_set_inner_vlan_id.vlan_id

    elif action.type == SX_FLEX_ACL_ACTION_SET_OUTER_VLAN_ID:
        action_fields = "VLAN: 0x%x" % action.fields.action_set_outer_vlan_id.vlan_id

    elif action.type == SX_FLEX_ACL_ACTION_SET_SRC_MAC:
        action_fields = "SMAC: %s" % action.fields.action_set_src_mac.mac.to_str()

    elif action.type == SX_FLEX_ACL_ACTION_SET_DST_MAC:
        action_fields = "SMAC: %s" % action.fields.action_set_dst_mac.mac.to_str()

    elif action.type == SX_FLEX_ACL_ACTION_SET_DSCP:
        action_fields = "DSCP: 0x%x" % action.fields.action_set_dscp.dscp_val

    elif action.type == SX_FLEX_ACL_ACTION_SET_BRIDGE:
        action_fields = "Bridge ID: 0x%x" % action.fields.action_set_bridge.bridge_id

    elif action.type == SX_FLEX_ACL_ACTION_PBS:
        action_fields = "PBS ID: 0x%x" % action.fields.action_pbs.pbs_id

    elif action.type == SX_FLEX_ACL_ACTION_SET_TC:
        action_fields = "TC: 0x%x" % action.fields.action_set_tc.tc_val

    elif action.type == SX_FLEX_ACL_ACTION_SET_TTL:
        action_fields = "TTL: 0x%x" % action.fields.action_set_ttl.ttl_val

    elif action.type == SX_FLEX_ACL_ACTION_DEC_TTL:
        action_fields = "TTL: 0x%x" % action.fields.action_dec_ttl.ttl_val

    elif action.type == SX_FLEX_ACL_ACTION_SET_COLOR:
        if action.fields.action_set_color.color_val in action_color_type_dict:
            action_fields = action_color_type_dict[action.fields.action_set_color.color_val]

    elif action.type == SX_FLEX_ACL_ACTION_SET_ECN:
        action_fields = "ECN: 0x%x" % action.fields.action_set_ecn.ecn_val

    elif action.type == SX_FLEX_ACL_ACTION_SET_USER_TOKEN:
        action_fields = "User Token: 0x%x, Mask: 0x%x" % (action.fields.action_set_user_token.user_token, action.fields.action_set_user_token.mask)

    elif action.type == SX_FLEX_ACL_ACTION_DONT_LEARN:
        action_fields = ""

    elif action.type == SX_FLEX_ACL_ACTION_RPF:
        if action.fields.action_rpf.rpf_action in rpf_action_dict:
            if action.fields.action_rpf.rpf_param.rpf_param_type == SX_ACL_FLEX_RPF_PARAM_TYPE_IRIF:
                action_fields = "RPF Action: %s, RPF RIF: 0x%x" % (rpf_action_dict[action.fields.action_rpf.rpf_action], action.fields.action_rpf.rpf_param.rpf_param_value.rpf_rif)
            else:
                action_fields = "RPF Action: %s, RPF RIF: 0x%x" % (rpf_action_dict[action.fields.action_rpf.rpf_action], action.fields.action_rpf.rpf_param.rpf_param_value.rpf_group)

    elif action.type == SX_FLEX_ACL_ACTION_MC_ROUTE:
        action_fields = "Egress MC Container ID: 0x%x" % action.fields.action_mc_route.egress_mc_container

    elif action.type == SX_FLEX_ACL_ACTION_TUNNEL_DECAP:
        action_fields = "Tunnel ID: 0x%x" % action.fields.action_tunnel_decap.tunnel_id

    elif action.type == SX_FLEX_ACL_ACTION_GOTO:
        if action.fields.action_goto.goto_action_cmd in action_goto_cmd_dict:
            action_fields = "CMD: %s, ACL Group ID: 0x%x" % (action_goto_cmd_dict[action.fields.action_goto.goto_action_cmd], action.fields.action_goto.acl_group_id)

    elif action.type == SX_FLEX_ACL_ACTION_SET_ROUTER:
        action_fields = "VRID: 0x%x" % action.fields.action_set_router.vrid

    elif action.type == SX_FLEX_ACL_ACTION_MC:
        action_fields = "MC Container ID: 0x%x" % action.fields.action_mc.mc_container_id

    elif action.type == SX_FLEX_ACL_ACTION_UC_ROUTE:
        if action.fields.action_uc_route.uc_route_type == SX_UC_ROUTE_TYPE_NEXT_HOP:
            action_fields = "UC Route Type: NEXT_HOP, ECMP ID: 0x%x" % action.fields.action_uc_route.uc_route_param.ecmp_id
        elif action.fields.action_uc_route.uc_route_type == SX_UC_ROUTE_TYPE_LOCAL:
            action_fields = "UC Route Type: LOCAL, Local Egress RIF: 0x%x" % action.fields.action_uc_route.uc_route_param.local_egress_rif
        elif action.fields.action_uc_route.uc_route_type == SX_UC_ROUTE_TYPE_IP2ME:
            action_fields = "UC Route Type: IP2ME"

    elif action.type == SX_FLEX_ACL_ACTION_SET_DSCP_REWRITE:
        if action.fields.action_set_dscp_rewrite.set_rewrite_cmd in action_rewrite_cmd_dict:
            action_fields = "Rewrite CMD: %s" % action_rewrite_cmd_dict[action.fields.action_set_dscp_rewrite.set_rewrite_cmd]

    elif action.type == SX_FLEX_ACL_ACTION_SET_PCP_REWRITE:
        if action.fields.action_set_pcp_rewrite.set_rewrite_cmd in action_rewrite_cmd_dict:
            action_fields = "Rewrite CMD: %s" % action_rewrite_cmd_dict[action.fields.action_set_pcp_rewrite.set_rewrite_cmd]

    elif action.type == SX_FLEX_ACL_ACTION_EGRESS_MIRROR:
        action_fields = "Session ID: 0x%x" % action.fields.action_egress_mirror.session_id

    elif action.type == SX_FLEX_ACL_ACTION_IGNORE_EGRESS_VLAN_FILTER:
        action_fields = ""

    elif action.type == SX_FLEX_ACL_ACTION_IGNORE_EGRESS_STP_FILTER:
        action_fields = ""

    elif action.type == SX_FLEX_ACL_ACTION_DISABLE_OVERLAY_LEARNING:
        action_fields = ""

    elif action.type == SX_FLEX_ACL_ACTION_PORT_FILTER:
        action_fields = "MC Container ID: 0x%x" % action.fields.action_port_filter.mc_container_id

    elif action.type == SX_FLEX_ACL_ACTION_SET_MPLS_TTL:
        action_fields = "TTL: 0x%x" % action.fields.action_set_mpls_ttl.ttl_val

    elif action.type == SX_FLEX_ACL_ACTION_DEC_MPLS_TTL:
        action_fields = "TTL: 0x%x" % action.fields.action_dec_mpls_ttl.ttl_val

    elif action.type == SX_FLEX_ACL_ACTION_SET_EXP:
        action_fields = "EXP: 0x%x" % action.fields.action_set_exp.exp_val

    elif action.type == SX_FLEX_ACL_ACTION_SET_EXP_REWRITE:
        if action.fields.action_set_exp_rewrite.set_rewrite_cmd in action_rewrite_cmd_dict:
            action_fields = "Rewrite CMD: %s" % action_rewrite_cmd_dict[action.fields.action_set_exp_rewrite.set_rewrite_cmd]

    elif action.type == SX_FLEX_ACL_ACTION_PBILM:
        action_fields = "PBILM ID: 0x%x" % action.fields.action_pbilm.pbilm_id

    elif action.type == SX_FLEX_ACL_ACTION_NVE_TUNNEL_ENCAP:
        action_fields = "Underlay DIP: %s, Tunnel ID: %d" % (ip_addr_to_str(action.fields.action_nve_tunnel_encap.underlay_dip), action.fields.action_nve_tunnel_encap.tunnel_id)

    elif action.type == SX_FLEX_ACL_ACTION_NVE_MC_TUNNEL_ENCAP:
        action_fields = "MC Container ID: %d" % (action.fields.action_nve_mc_tunnel_encap.mc_container_id)

    elif action.type == SX_FLEX_ACL_ACTION_TRAP_W_USER_ID:
        if action.fields.action_trap_w_user_id.action in trap_action_dict:
            action_fields = "Action: %s, Trap ID: %d, User ID: %d" % (trap_action_dict[action.fields.action_trap_w_user_id.action], action.fields.action_trap_w_user_id.trap_id, action.fields.action_trap_w_user_id.user_id)

    return action_fields


def get_port_lag_bindings(log_port, binding_map):
    acl_id_p = new_sx_acl_id_t_p()
    rc = sx_api_acl_port_bind_get(handle, log_port, SX_ACL_DIRECTION_INGRESS, acl_id_p)
    if rc != SX_STATUS_SUCCESS and rc != SX_STATUS_ENTRY_NOT_FOUND:
        print ("sx_api_acl_port_bind_get failed, rc = %d" % (rc))
        exit()

    if rc == SX_STATUS_SUCCESS:
        acl_id = sx_acl_id_t_p_value(acl_id_p)
        if acl_id not in binding_map:
            binding_map[acl_id] = []
        binding_map[acl_id].append({'direction': SX_ACL_DIRECTION_INGRESS, 'log_port': log_port})

    rc = sx_api_acl_port_bind_get(handle, log_port, SX_ACL_DIRECTION_EGRESS, acl_id_p)
    if rc != SX_STATUS_SUCCESS and rc != SX_STATUS_ENTRY_NOT_FOUND:
        print ("sx_api_acl_port_bind_get failed, rc = %d" % (rc))
        exit()

    if rc == SX_STATUS_SUCCESS:
        acl_id = sx_acl_id_t_p_value(acl_id_p)
        if acl_id not in binding_map:
            binding_map[acl_id] = []
        binding_map[acl_id].append({'direction': SX_ACL_DIRECTION_EGRESS, 'log_port': log_port})


def get_all_port_lag_bindings(binding_map):
    device_info_cnt_p = new_uint32_t_p()
    uint32_t_p_assign(device_info_cnt_p, 0)
    rc = sx_api_port_device_list_get(handle, None, device_info_cnt_p)
    if rc != SX_STATUS_SUCCESS:
        print ("sx_api_port_device_list_get failed, rc = %d" % (rc))
        exit()
    device_info_cnt = uint32_t_p_value(device_info_cnt_p)
    device_info_list_p = new_sx_device_info_t_arr(device_info_cnt)
    rc = sx_api_port_device_list_get(handle, device_info_list_p, device_info_cnt_p)
    if rc != SX_STATUS_SUCCESS:
        print ("sx_api_port_device_list_get failed, rc = %d" % (rc))
        exit()
    device_info_cnt = uint32_t_p_value(device_info_cnt_p)

    swid_cnt_p = new_uint32_t_p()
    uint32_t_p_assign(swid_cnt_p, 0)
    rc = sx_api_port_swid_list_get(handle, None, swid_cnt_p)
    if rc != SX_STATUS_SUCCESS:
        print ("sx_api_port_swid_list_get failed, rc = %d" % (rc))
        exit()
    swid_cnt = uint32_t_p_value(swid_cnt_p)
    swid_list_p = new_sx_swid_t_arr(swid_cnt)
    rc = sx_api_port_swid_list_get(handle, swid_list_p, swid_cnt_p)
    if rc != SX_STATUS_SUCCESS:
        print ("sx_api_port_swid_list_get failed, rc = %d" % (rc))
        exit()
    swid_cnt = uint32_t_p_value(swid_cnt_p)

    port_id_list = []
    lag_id_list = []
    lag_member_port_list = []

    for i in range(0, device_info_cnt):
        device_info = sx_device_info_t_arr_getitem(device_info_list_p, i)
        device_id = device_info.dev_id
        for j in range(0, swid_cnt):
            swid = sx_swid_t_arr_getitem(swid_list_p, j)
            port_cnt_p = new_uint32_t_p()
            uint32_t_p_assign(port_cnt_p, 0)
            rc = sx_api_port_device_get(handle, device_id, swid, None, port_cnt_p)
            if rc != SX_STATUS_SUCCESS:
                print ("sx_api_port_device_get failed, rc = %d" % (rc))
                exit()
            port_cnt = uint32_t_p_value(port_cnt_p)
            port_attributes_list = new_sx_port_attributes_t_arr(port_cnt)
            rc = sx_api_port_device_get(handle, device_id, swid, port_attributes_list, port_cnt_p)
            if rc != SX_STATUS_SUCCESS:
                print ("sx_api_port_device_get failed, rc = %d" % (rc))
                exit()
            port_cnt = uint32_t_p_value(port_cnt_p)
            for k in range(0, port_cnt):
                port_attributes = sx_port_attributes_t_arr_getitem(port_attributes_list, k)
                log_port = port_attributes.log_port
                port_id_list.append(log_port)

            lag_id_cnt_p = new_uint32_t_p()
            uint32_t_p_assign(lag_id_cnt_p, 0)
            rc = sx_api_lag_port_group_iter_get(handle, SX_ACCESS_CMD_GET, swid, 0, None, None, lag_id_cnt_p)
            if rc != SX_STATUS_SUCCESS:
                print ("sx_api_lag_port_group_iter_get failed, rc = %d" % (rc))
                exit()
            lag_id_cnt = uint32_t_p_value(lag_id_cnt_p)
            lag_id_list_p = new_sx_port_log_id_t_arr(lag_id_cnt)
            rc = sx_api_lag_port_group_iter_get(handle, SX_ACCESS_CMD_GET_FIRST, swid, 0, None, lag_id_list_p, lag_id_cnt_p)
            if rc != SX_STATUS_SUCCESS:
                print ("sx_api_lag_port_group_iter_get failed, rc = %d" % (rc))
                exit()
            lag_id_cnt = uint32_t_p_value(lag_id_cnt_p)
            for k in range(0, lag_id_cnt):
                lag_id = sx_port_log_id_t_arr_getitem(lag_id_list_p, k)
                lag_id_list.append(lag_id)
                log_port_cnt_p = new_uint32_t_p()
                uint32_t_p_assign(log_port_cnt_p, 0)
                rc = sx_api_lag_port_group_get(handle, swid, lag_id, None, log_port_cnt_p)
                if rc != SX_STATUS_SUCCESS:
                    print ("sx_api_lag_port_group_get failed, rc = %d" % (rc))
                    exit()
                log_port_cnt = uint32_t_p_value(log_port_cnt_p)
                if log_port_cnt == 0:
                    continue
                log_port_list_p = new_sx_port_log_id_t_arr(log_port_cnt)
                rc = sx_api_lag_port_group_get(handle, swid, lag_id, log_port_list_p, log_port_cnt_p)
                if rc != SX_STATUS_SUCCESS:
                    print ("sx_api_lag_port_group_get failed, rc = %d" % (rc))
                    exit()
                log_port_cnt = uint32_t_p_value(log_port_cnt_p)
                for n in range(0, log_port_cnt):
                    log_port = sx_port_log_id_t_arr_getitem(log_port_list_p, n)
                    lag_member_port_list.append(log_port)

    for member_port in lag_member_port_list:
        if member_port in port_id_list:
            port_id_list.remove(member_port)

    for port_id in port_id_list:
        get_port_lag_bindings(port_id, binding_map)

    for lag_id in lag_id_list:
        get_port_lag_bindings(lag_id, binding_map)


def get_all_rif_bindings(binding_map):
    rc = sx_api_router_log_verbosity_level_set(handle, SX_LOG_VERBOSITY_BOTH, SX_VERBOSITY_LEVEL_NONE, SX_VERBOSITY_LEVEL_NONE)
    if rc != SX_STATUS_SUCCESS:
        print ("fail to set router API log verbosity level")
        exit()
    rif_cnt_p = new_uint32_t_p()
    uint32_t_p_assign(rif_cnt_p, 0)
    rc = sx_api_router_interface_iter_get(handle, SX_ACCESS_CMD_GET, None, None, None, rif_cnt_p)
    if rc == SX_STATUS_MODULE_UNINITIALIZED:
        return
    elif rc != SX_STATUS_SUCCESS:
        print ("sx_api_router_interface_iter_get failed, rc = %d" % (rc))
        exit()
    rif_cnt = uint32_t_p_value(rif_cnt_p)
    rif_list_p = new_sx_router_interface_t_arr(rif_cnt)
    rc = sx_api_router_interface_iter_get(handle, SX_ACCESS_CMD_GET_FIRST, None, None, rif_list_p, rif_cnt_p)
    if rc != SX_STATUS_SUCCESS:
        print ("sx_api_router_interface_iter_get failed, rc = %d" % (rc))
        exit()
    rif_cnt = uint32_t_p_value(rif_cnt_p)
    for i in xrange(0, rif_cnt):
        rif = sx_router_interface_t_arr_getitem(rif_list_p, i)
        acl_id_p = new_sx_acl_id_t_p()
        rc = sx_api_acl_rif_bind_get(handle, rif, SX_ACL_DIRECTION_RIF_INGRESS, acl_id_p)
        if rc != SX_STATUS_SUCCESS and rc != SX_STATUS_ENTRY_NOT_FOUND:
            print ("sx_api_acl_rif_bind_get failed, rc = %d" % (rc))
            exit()

        if rc == SX_STATUS_SUCCESS:
            acl_id = sx_acl_id_t_p_value(acl_id_p)
            if acl_id not in binding_map:
                binding_map[acl_id] = []
            binding_map[acl_id].append({'direction': SX_ACL_DIRECTION_RIF_INGRESS, 'rif': rif})

        rc = sx_api_acl_rif_bind_get(handle, rif, SX_ACL_DIRECTION_RIF_EGRESS, acl_id_p)
        if rc != SX_STATUS_SUCCESS and rc != SX_STATUS_ENTRY_NOT_FOUND:
            print ("sx_api_acl_rif_bind_get failed, rc = %d" % (rc))
            exit()

        if rc == SX_STATUS_SUCCESS:
            acl_id = sx_acl_id_t_p_value(acl_id_p)
            if acl_id not in binding_map:
                binding_map[acl_id] = []
            binding_map[acl_id].append({'direction': SX_ACL_DIRECTION_RIF_EGRESS, 'rif': rif})


def get_all_vlan_group_bindings(binding_map):
    rc = sx_api_acl_log_verbosity_level_set(handle, SX_LOG_VERBOSITY_BOTH, SX_VERBOSITY_LEVEL_NONE, SX_VERBOSITY_LEVEL_NONE)
    if rc != SX_STATUS_SUCCESS:
        print ("fail to set ACL API log verbosity level")
        exit()
    VLAN_GROUP_MAX = 255
    for vlan_group in xrange(0, VLAN_GROUP_MAX):
        acl_id_p = new_sx_acl_id_t_p()
        rc = sx_api_acl_vlan_group_bind_get(handle, vlan_group, SX_ACL_DIRECTION_INGRESS, acl_id_p)
        if rc != SX_STATUS_SUCCESS and rc != SX_STATUS_ENTRY_NOT_FOUND and rc != SX_STATUS_PARAM_EXCEEDS_RANGE:
            print ("sx_api_acl_vlan_group_bind_get failed, rc = %d" % (rc))
            exit()

        if rc == SX_STATUS_PARAM_EXCEEDS_RANGE:
            break

        if rc == SX_STATUS_SUCCESS:
            acl_id = sx_acl_id_t_p_value(acl_id_p)
            if acl_id not in binding_map:
                binding_map[acl_id] = []
            binding_map[acl_id].append({'direction': SX_ACL_DIRECTION_INGRESS, 'vlan_group': vlan_group})

        rc = sx_api_acl_vlan_group_bind_get(handle, vlan_group, SX_ACL_DIRECTION_EGRESS, acl_id_p)
        if rc != SX_STATUS_SUCCESS and rc != SX_STATUS_ENTRY_NOT_FOUND and rc != SX_STATUS_PARAM_EXCEEDS_RANGE:
            print ("sx_api_acl_vlan_group_bind_get failed, rc = %d" % (rc))
            exit()

        if rc == SX_STATUS_SUCCESS:
            acl_id = sx_acl_id_t_p_value(acl_id_p)
            if acl_id not in binding_map:
                binding_map[acl_id] = []
            binding_map[acl_id].append({'direction': SX_ACL_DIRECTION_EGRESS, 'vlan_group': vlan_group})


old_stdout = redirect_stdout()
rc, handle = sx_api_open(None)
sys.stdout = os.fdopen(old_stdout, 'w')

acl_id_list = []
acl_group_id_cnt_p = new_uint32_t_p()
acl_direction_p = new_sx_acl_direction_t_p()
uint32_t_p_assign(acl_group_id_cnt_p, 0)
rc = sx_api_acl_group_iter_get(handle, SX_ACCESS_CMD_GET, 0, None, None, acl_group_id_cnt_p)
if rc != SX_STATUS_SUCCESS:
    print ("sx_api_acl_group_iter_get failed, rc = %d" % (rc))
    exit()
acl_group_id_cnt = uint32_t_p_value(acl_group_id_cnt_p)
acl_group_id_list_p = new_sx_acl_id_t_arr(acl_group_id_cnt)
rc = sx_api_acl_group_iter_get(handle, SX_ACCESS_CMD_GET_FIRST, 0, None, acl_group_id_list_p, acl_group_id_cnt_p)
if rc != SX_STATUS_SUCCESS:
    print ("sx_api_acl_group_iter_get failed, rc = %d" % (rc))
    exit()
acl_group_id_cnt = uint32_t_p_value(acl_group_id_cnt_p)
acl_group_header_printed = False

prev_group_dict = {}
next_group_dict = {}

next_group_id_p = new_uint32_t_p()
for i in xrange(0, acl_group_id_cnt):
    acl_group_id = sx_acl_id_t_arr_getitem(acl_group_id_list_p, i)
    acl_id_list.append(acl_group_id)
    if acl_group_id not in prev_group_dict:
        prev_group_dict[acl_group_id] = INVALID_GROUP_ID
    if acl_group_id not in next_group_dict:
        next_group_dict[acl_group_id] = INVALID_GROUP_ID
    rc = sx_api_acl_group_bind_get(handle, acl_group_id, next_group_id_p)
    if rc != SX_STATUS_SUCCESS:
        print ("sx_api_acl_group_bind_get failed, rc = %d" % (rc))
        exit()
    next_group_id = uint32_t_p_value(next_group_id_p)
    next_group_dict[acl_group_id] = next_group_id
    prev_group_dict[next_group_id] = acl_group_id

region_set = set()
for i in xrange(0, acl_group_id_cnt):
    if not acl_group_header_printed:
        print ("----------------------------")
        print ("ACL Groups")
        print ("----------------------------\n")

        print ("========================================================================")
        print ("|%10s|%12s|%11s|%11s|%11s|%10s|" % ("Group ID", "Direction", "Prev Group", "Next Group", "ACL ID", "Region ID"))
        print ("========================================================================")
        acl_group_header_printed = True

    acl_group_id = sx_acl_id_t_arr_getitem(acl_group_id_list_p, i)
    acl_id_cnt_p = new_uint32_t_p()
    uint32_t_p_assign(acl_id_cnt_p, 0)
    rc = sx_api_acl_group_get(handle, acl_group_id, acl_direction_p, None, acl_id_cnt_p)
    if rc != SX_STATUS_SUCCESS:
        print ("sx_api_acl_group_get failed, rc = %d" % (rc))
        exit()
    acl_id_cnt = uint32_t_p_value(acl_id_cnt_p)

    acl_id_list_p = new_sx_acl_id_t_arr(acl_id_cnt)
    rc = sx_api_acl_group_get(handle, acl_group_id, acl_direction_p, acl_id_list_p, acl_id_cnt_p)
    if rc != SX_STATUS_SUCCESS:
        print ("sx_api_acl_group_get failed, rc = %d" % (rc))
        exit()
    acl_id_cnt = uint32_t_p_value(acl_id_cnt_p)

    print ("|%10s %12s %11s %11s|" % ("0x%x" % acl_group_id,
                                      dir_dict[sx_acl_direction_t_p_value(acl_direction_p)],
                                      "0x%x" % prev_group_dict[acl_group_id],
                                      "0x%x" % next_group_dict[acl_group_id]), end='')

    if acl_id_cnt == 0:
        print ("%11s|%10s|" % ("", ""))
        print ("========================================================================")
    else:
        acl_region_pair_list = []

        for j in xrange(0, acl_id_cnt):
            acl_id = sx_acl_id_t_arr_getitem(acl_id_list_p, j)
            acl_type_p = new_sx_acl_type_t_p()
            acl_region_group_p = new_sx_acl_region_group_t_p()
            rc = sx_api_acl_get(handle, acl_id, acl_type_p, acl_direction_p, acl_region_group_p)
            if rc != SX_STATUS_SUCCESS:
                print ("sx_api_acl_get failed, rc = %d" % (rc))
                exit()
            acl_region_group = sx_acl_region_group_t_p_value(acl_region_group_p)
            if acl_region_group.acl_type == SX_ACL_TYPE_PACKET_TYPES_AGNOSTIC:
                acl_region_pair_list.append((acl_id, acl_region_group.regions.acl_packet_agnostic.region))
                region_set.add(acl_region_group.regions.acl_packet_agnostic.region)
            else:
                acl_region_pair_list.append((acl_id, acl_region_group.regions.acl_packet_sensitive.non_ip_region))
                acl_region_pair_list.append((acl_id, acl_region_group.regions.acl_packet_sensitive.ipv4_region))
                acl_region_pair_list.append((acl_id, acl_region_group.regions.acl_packet_sensitive.ipv6_region))
                region_set.add(acl_region_group.regions.acl_packet_sensitive.non_ip_region)
                region_set.add(acl_region_group.regions.acl_packet_sensitive.ipv4_region)
                region_set.add(acl_region_group.regions.acl_packet_sensitive.ipv6_region)

        for k in xrange(0, len(acl_region_pair_list)):
            if k == 0:
                print ("%11s|%10s|" % ("0x%x" % acl_region_pair_list[k][0], "0x%x" % acl_region_pair_list[k][1]))
            else:
                print ("|%47s|%11s|%10s|" % ("", "0x%x" % acl_region_pair_list[k][0], "0x%x" % acl_region_pair_list[k][1]))

            if k == (len(acl_region_pair_list) - 1):
                print ("========================================================================")
            else:
                print ("|%47s|-----------------------" % (""))

acl_id_cnt_p = new_uint32_t_p()
uint32_t_p_assign(acl_id_cnt_p, 0)
rc = sx_api_acl_iter_get(handle, SX_ACCESS_CMD_GET, 0, None, None, acl_id_cnt_p)
if rc != SX_STATUS_SUCCESS:
    print ("sx_api_acl_iter_get failed, rc = %d" % (rc))
    exit()

acl_id_cnt = uint32_t_p_value(acl_id_cnt_p)
acl_id_list_p = new_sx_acl_id_t_arr(acl_id_cnt)
rc = sx_api_acl_iter_get(handle, SX_ACCESS_CMD_GET_FIRST, 0, None, acl_id_list_p, acl_id_cnt_p)
if rc != SX_STATUS_SUCCESS:
    print ("sx_api_acl_iter_get failed, rc = %d" % (rc))
    exit()
acl_id_cnt = uint32_t_p_value(acl_id_cnt_p)
for i in xrange(0, acl_id_cnt):
    acl_id = sx_acl_id_t_arr_getitem(acl_id_list_p, i)
    acl_id_list.append(acl_id)
    acl_type_p = new_sx_acl_type_t_p()
    acl_region_group_p = new_sx_acl_region_group_t_p()
    rc = sx_api_acl_get(handle, acl_id, acl_type_p, acl_direction_p, acl_region_group_p)
    if rc != SX_STATUS_SUCCESS:
        print ("sx_api_acl_get failed, rc = %d" % (rc))
        exit()
    acl_region_pair_list = []
    acl_region_group = sx_acl_region_group_t_p_value(acl_region_group_p)
    if acl_region_group.acl_type == SX_ACL_TYPE_PACKET_TYPES_AGNOSTIC:
        acl_region_pair_list.append((acl_id, acl_region_group.regions.acl_packet_agnostic.region))
        region_set.add(acl_region_group.regions.acl_packet_agnostic.region)
    else:
        acl_region_pair_list.append((acl_id, acl_region_group.regions.acl_packet_sensitive.non_ip_region))
        acl_region_pair_list.append((acl_id, acl_region_group.regions.acl_packet_sensitive.ipv4_region))
        acl_region_pair_list.append((acl_id, acl_region_group.regions.acl_packet_sensitive.ipv6_region))
        region_set.add(acl_region_group.regions.acl_packet_sensitive.non_ip_region)
        region_set.add(acl_region_group.regions.acl_packet_sensitive.ipv4_region)
        region_set.add(acl_region_group.regions.acl_packet_sensitive.ipv6_region)

region_list = sorted(region_set)
acl_rule_header_printed = False
for i in xrange(0, len(region_list)):
    if not acl_rule_header_printed:
        print ("\n----------------------------")
        print ("ACL Rules")
        print ("----------------------------\n")

        print ("========================================================================================================================================================================================================================================")
        print ("|%10s|%8s|%30s|%40s|%40s|%26s|%70s|" % ("Region ID", "Offset", "Key Type", "Key Value", "Key Mask", "Action Type", "Action Fields"))
        print ("========================================================================================================================================================================================================================================")
        acl_rule_header_printed = True

    region_id = region_list[i]

    rules_cnt_p = new_uint32_t_p()
    uint32_t_p_assign(rules_cnt_p, 0)
    rc = sx_api_acl_flex_rules_get(handle, region_id, None, None, rules_cnt_p)
    if rc != SX_STATUS_SUCCESS:
        print ("sx_api_acl_flex_rules_get failed, rc = %d" % (rc))
        exit()
    rules_cnt = uint32_t_p_value(rules_cnt_p)

    offsets_list_p = new_sx_acl_rule_offset_t_arr(rules_cnt)
    for j in xrange(0, rules_cnt):
        sx_acl_rule_offset_t_arr_setitem(offsets_list_p, j, 0)

    rules_list_p = new_sx_flex_acl_flex_rule_t_arr(rules_cnt)
    for j in xrange(0, rules_cnt):
        rule = sx_flex_acl_flex_rule_t()
        rule.key_desc_list_p = None
        rule.action_list_p = None
        sx_flex_acl_flex_rule_t_arr_setitem(rules_list_p, j, rule)

    rc = sx_api_acl_flex_rules_get(handle, region_id, offsets_list_p, rules_list_p, rules_cnt_p)
    if rc != SX_STATUS_SUCCESS:
        print ("sx_api_acl_flex_rules_get failed, rc = %d" % (rc))
        exit()
    rules_cnt = uint32_t_p_value(rules_cnt_p)

    for j in xrange(0, rules_cnt):
        rule = sx_flex_acl_flex_rule_t_arr_getitem(rules_list_p, j)
        rule.key_desc_list_p = new_sx_flex_acl_key_desc_t_arr(rule.key_desc_count)
        rule.action_list_p = new_sx_flex_acl_flex_action_t_arr(rule.action_count)
        sx_flex_acl_flex_rule_t_arr_setitem(rules_list_p, j, rule)

    rc = sx_api_acl_flex_rules_get(handle, region_id, offsets_list_p, rules_list_p, rules_cnt_p)
    if rc != SX_STATUS_SUCCESS:
        print ("sx_api_acl_flex_rules_get failed, rc = %d" % (rc))
        exit()
    rules_cnt = uint32_t_p_value(rules_cnt_p)

    print ("|%10s|" % ("0x%x" % region_id), end='')

    if rules_cnt == 0:
        print ("%8s|%30s|%40s|%40s|%26s|%70s|" % ("", "", "", "", "", ""))
        print ("========================================================================================================================================================================================================================================")
    else:
        rules_list = []
        for j in xrange(0, rules_cnt):
            offset = sx_acl_rule_offset_t_arr_getitem(offsets_list_p, j)
            rule = sx_flex_acl_flex_rule_t_arr_getitem(rules_list_p, j)
            key_cnt = rule.key_desc_count
            action_cnt = rule.action_count
            max_cnt = max(key_cnt, action_cnt)
            if max_cnt == 0:
                rules_list.append(("%d" % offset, "", "", "", "", ""))
            else:
                for k in xrange(0, max_cnt):
                    key_type = ""
                    key_value = ""
                    key_mask = ""
                    action_type = ""
                    action_fields = ""
                    if k < key_cnt and k >= action_cnt:
                        key_desc = sx_flex_acl_key_desc_t_arr_getitem(rule.key_desc_list_p, k)
                        if key_desc.key_id in key_type_dict:
                            key_type = key_type_dict[key_desc.key_id]
                        else:
                            key_type = "Unknown"

                        key_value, key_mask = get_key_val_mask_str(key_desc)
                    elif k >= key_cnt and k < action_cnt:
                        action = sx_flex_acl_flex_action_t_arr_getitem(rule.action_list_p, k)
                        if action.type in action_type_dict:
                            action_type = action_type_dict[action.type]
                        else:
                            action_type = "Unknown"

                        action_fields = get_action_fields_str(action)
                    else:
                        key_desc = sx_flex_acl_key_desc_t_arr_getitem(rule.key_desc_list_p, k)
                        if key_desc.key_id in key_type_dict:
                            key_type = key_type_dict[key_desc.key_id]
                        else:
                            key_type = "Unknown"
                        action = sx_flex_acl_flex_action_t_arr_getitem(rule.action_list_p, k)
                        if action.type in action_type_dict:
                            action_type = action_type_dict[action.type]
                        else:
                            action_type = "Unknown"

                        key_value, key_mask = get_key_val_mask_str(key_desc)
                        action_fields = get_action_fields_str(action)

                    rules_list.append(("%d" % offset, key_type, key_value, key_mask, action_type, action_fields))

        rules_list_len = len(rules_list)
        for j in xrange(0, rules_list_len):
            if j == 0:
                print ("%8s|%30s|%40s|%40s|%26s|%70s|" % rules_list[j])
            else:
                print ("|%10s|%8s|%30s|%40s|%40s|%26s|%70s|" % ("", rules_list[j][0], rules_list[j][1], rules_list[j][2], rules_list[j][3], rules_list[j][4], rules_list[j][5]))

            if j == (rules_list_len - 1):
                print ("========================================================================================================================================================================================================================================")
            else:
                if rules_list[j][0] != rules_list[j + 1][0]:
                    print ("|%10s|============================================================================================================================================================================================================================" % (""))
                else:
                    print ("|%10s|----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------" % (""))

binding_map = {}
get_all_port_lag_bindings(binding_map)
get_all_rif_bindings(binding_map)
get_all_vlan_group_bindings(binding_map)
acl_binding_header_printed = False
for i in xrange(0, len(acl_id_list)):
    if not acl_binding_header_printed:
        print ("\n----------------------------")
        print ("ACL Bindings")
        print ("----------------------------\n")

        print ("==========================================================")
        print ("|%11s|%12s|%13s|%11s|%5s|" % ("ACL ID", "Direction", "Logical Port", "VLAN Group", "RIF"))
        print ("==========================================================")
        acl_binding_header_printed = True

    acl_id = acl_id_list[i]
    if acl_id in binding_map:
        acl_id_str = "0x%x" % (acl_id)
        print ("|%11s|" % (acl_id_str), end='')
        for j in xrange(0, len(binding_map[acl_id])):
            entry = binding_map[acl_id][j]
            dir_str = ""
            log_port_str = ""
            vlan_group_str = ""
            rif_str = ""
            if 'direction' in entry and entry['direction'] in dir_dict:
                dir_str = dir_dict[entry['direction']]
            if 'log_port' in entry:
                log_port_str = "0x%x" % (entry['log_port'])
            if 'vlan_group' in entry:
                vlan_group_str = "%d" % (entry['vlan_group'])
            if 'rif' in entry:
                rif_str = "%d" % (entry['rif'])

            if j == 0:
                print ("%12s|%13s|%11s|%5s|" % (dir_str, log_port_str, vlan_group_str, rif_str))
            else:
                print ("|%11s|%12s|%13s|%11s|%5s|" % ("", dir_str, log_port_str, vlan_group_str, rif_str))

            if j == (len(binding_map[acl_id]) - 1):
                print ("==========================================================")
            else:
                print ("|%11s|---------------------------------------------" % (""))
