#!/usr/bin/env python
"""
This file contains Python command example for the MC ROUTER module.
Python commands syntax is very similar to the Switch SDK APIs. 
You can learn more about each command and its parameters by reading the SwitchX API Reference Manual.
"""
import sys, socket, struct
from argparse import Action

from python_sdk_api.sx_api import *
import sx_api_router as sx_api_router
import sys
from test_infra_common import * 

port_list = mapPortAndInterfaces(sx_api_router.handle)
PORT1 = port_list[0]
PORT2 = port_list[1]
PORT3 = port_list[2]
PORT4 = port_list[3]
DONT_CARE_RIF = 1001

def prepare_vlan_rif(vrid, port, vlan, mac, mtu = 1500):
    sx_api_router.add_ports_to_vlan(vlan, {port: SX_TAGGED_MEMBER})
    rif = sx_api_router.create_vlan_rif(vrid, vlan, mac, mtu)
    return rif
    
def destroy_vlan_rif(vrid, port, vlan, rif):
    sx_api_router.delete_rif(vrid, rif)
    sx_api_router.remove_ports_from_vlan(vlan, {port: SX_TAGGED_MEMBER})

def prepare_bridge_rif(vrid, port_vlan_list, mac, mtu = 1500):
    bridge_id = sx_api_router.create_bridge()
    
    vport_arr = [None] * len(port_vlan_list)
    for i, port_vlan in enumerate(port_vlan_list):
        sx_api_router.add_ports_to_vlan(port_vlan[1], {port_vlan[0]: SX_TAGGED_MEMBER})
        vport_arr[i] = sx_api_router.create_vport(port_vlan[0], port_vlan[1], SX_TAGGED_MEMBER)
        sx_api_router.add_vport_to_bridge(bridge_id, vport_arr[i])
        sx_api_router.set_port_state(vport_arr[i], SX_PORT_ADMIN_STATUS_UP)
        
    rif = sx_api_router.create_bridge_rif(vrid, bridge_id, mac, mtu)
    return vport_arr, bridge_id, rif

def destroy_bridge_rif(vrid, vport_list, bridge_id, rif, port_vlan_list):
    sx_api_router.delete_rif(vrid, rif)
    for i, vport in enumerate(vport_list):
        sx_api_router.set_port_state(vport, SX_PORT_ADMIN_STATUS_DOWN)
        sx_api_router.delete_vport_from_bridge(bridge_id, vport)
        sx_api_router.delete_vport(port_vlan_list[i][0], vport, port_vlan_list[i][1], SX_TAGGED_MEMBER)
        
    sx_api_router.destroy_bridge(bridge_id)
    
def prepare_router_port_rif(vrid, port, mac, mtu = 1500):
    rif = sx_api_router.create_router_port_rif(vrid, port, mac, mtu)
    return rif
    
def destroy_router_port_rif(vrid, rif):
    sx_api_router.delete_rif(vrid, rif)
        
def prepare_vport_rif(vrid, port, vlan, mac, mtu = 1500):
    sx_api_router.add_ports_to_vlan(vlan, {port: SX_TAGGED_MEMBER})
    vport = sx_api_router.create_vport(port, vlan, SX_TAGGED_MEMBER)
    rif = sx_api_router.create_vport_rif(vrid, vport, mac, mtu)
    return vport, rif

def destroy_vport_rif(vrid, port, vlan, vport, rif):
    sx_api_router.delete_rif(vrid, rif)
    sx_api_router.delete_vport(port, vport, vlan, SX_TAGGED_MEMBER)
    sx_api_router.remove_ports_from_vlan(vlan, {port: SX_TAGGED_MEMBER})
        
def make_ipv4_mc_route_key(ingress_rif, s_addr, s_mask, g_addr, g_mask):
    mc_route_key = sx_mc_route_key_t()
    mc_route_key.source_addr = sx_api_router.make_sx_ip_prefix_v4(s_addr, s_mask)
    mc_route_key.mc_group_addr = sx_api_router.make_sx_ip_prefix_v4(g_addr, g_mask)
    mc_route_key.ingress_rif = ingress_rif
    return mc_route_key

def make_mc_route_attributes(rpf_action = SX_ROUTER_RPF_ACTION_DISABLED,
                               ttl_cmd = SX_ROUTE_TTL_CMD_DEC,
                               ttl_val = 0,
                               prio_type = SX_PRIO_TYPE_MANUAL,
                               manual_prio = 1,
                               rpf_group_id = 0):
    mc_attr = sx_mc_route_attributes_t()
    mc_attr.rpf_action = rpf_action
    mc_attr.ttl_cmd = ttl_cmd
    mc_attr.ttl_val = ttl_val
    mc_attr.prio.type = prio_type
    mc_attr.prio.manual_prio = manual_prio
    mc_attr.rpf_group_id = rpf_group_id
    return mc_attr
    
def make_egress_rif_list(erif_list):
    erif_cnt = len(erif_list)
    erif_arr = new_sx_router_interface_t_arr(erif_cnt)
    for i, erif in enumerate(erif_list):
        sx_router_interface_t_arr_setitem(erif_arr, i, erif)
    return erif_arr

def make_mc_route_data(erif_list,
                       action = SX_ROUTER_ACTION_FORWARD,
                       trap_prio = 0):
    mc_route_data = sx_mc_route_data_t()
    mc_route_data.action = action
    mc_route_data.trap_attr.prio = trap_prio
    mc_route_data.egress_rif_cnt = len(erif_list)
    for i, erif in enumerate(erif_list):
        sx_router_interface_t_arr_setitem(mc_route_data.egress_rif_list_p, i, erif)
    return mc_route_data
   
def add_mc_route(vrid, mc_route_key, mc_route_attr, mc_route_data):
    mc_route_key_p = new_sx_mc_route_key_t_p()
    sx_mc_route_key_t_p_assign(mc_route_key_p, mc_route_key)
    
    mc_route_attr_p = new_sx_mc_route_attributes_t_p()
    sx_mc_route_attributes_t_p_assign(mc_route_attr_p, mc_route_attr)
    
    mc_route_data_p = new_sx_mc_route_data_t_p()
    sx_mc_route_data_t_p_assign(mc_route_data_p, mc_route_data)
    
    rc = sx_api_router_mc_route_set(sx_api_router.handle, 
                                    SX_ACCESS_CMD_ADD,
                                    vrid,
                                    mc_route_key_p,
                                    mc_route_attr_p,
                                    mc_route_data_p)
    assert SX_STATUS_SUCCESS == rc, "Failed to add MC route to VRID %d, rc: %d" % (vrid, rc)
    print "Added MC route to VRID %d" % (vrid)
    
def modify_mc_route(vrid, mc_route_key, mc_route_attr, mc_route_data):
    mc_route_key_p = new_sx_mc_route_key_t_p()
    sx_mc_route_key_t_p_assign(mc_route_key_p, mc_route_key)
    
    mc_route_attr_p = new_sx_mc_route_attributes_t_p()
    sx_mc_route_attributes_t_p_assign(mc_route_attr_p, mc_route_attr)
    
    mc_route_data_p = new_sx_mc_route_data_t_p()
    sx_mc_route_data_t_p_assign(mc_route_data_p, mc_route_data)
    
    rc = sx_api_router_mc_route_set(sx_api_router.handle, 
                                    SX_ACCESS_CMD_EDIT,
                                    vrid,
                                    mc_route_key_p,
                                    mc_route_attr_p,
                                    mc_route_data_p)
    assert SX_STATUS_SUCCESS == rc, "Failed to modify MC route to VRID %d, rc: %d" % (vrid, rc)
    print "Modified MC route in VRID %d" % (vrid)
    
def delete_mc_route(vrid, mc_route_key):
    mc_route_key_p = new_sx_mc_route_key_t_p()
    sx_mc_route_key_t_p_assign(mc_route_key_p, mc_route_key)
    
    rc = sx_api_router_mc_route_set(sx_api_router.handle, 
                                    SX_ACCESS_CMD_DELETE,
                                    vrid,
                                    mc_route_key_p,
                                    None,
                                    None)
    assert SX_STATUS_SUCCESS == rc, "Failed to delete MC route from VRID %d, rc: %d" % (vrid, rc)
    print "Deleted MC route from VRID %d" % (vrid)
    
def delete_all_mc_routes(vrid, protocol = SX_IP_VERSION_NONE):
    if (protocol == SX_IP_VERSION_IPV4):
        mc_route_key = make_ipv4_mc_route_key(0, '0.0.0.0', '0.0.0.0', '0.0.0.0', '0.0.0.0')
        mc_route_key_p = new_sx_mc_route_key_t_p()
        sx_mc_route_key_t_p_assign(mc_route_key_p, mc_route_key)
    else:
        mc_route_key_p = None
        
    rc = sx_api_router_mc_route_set(sx_api_router.handle, 
                                    SX_ACCESS_CMD_DELETE_ALL,
                                    vrid,
                                    mc_route_key_p,
                                    None,
                                    None)
    assert SX_STATUS_SUCCESS == rc, "Failed to delete all MC routes from VRID %d, rc: %d" % (vrid, rc)
    print "Deleted all MC routes from VRID %d" % (vrid)
    
def get_mc_route_activity(vrid, mc_route_key, clear = False):
    mc_route_key_p = new_sx_mc_route_key_t_p()
    sx_mc_route_key_t_p_assign(mc_route_key_p, mc_route_key)
    
    activity_p = new_boolean_t_p()
    
    cmd = SX_ACCESS_CMD_READ_CLEAR if clear else SX_ACCESS_CMD_READ
    
    rc = sx_api_router_mc_route_activity_get(sx_api_router.handle,
                                             cmd,
                                             vrid,
                                             mc_route_key_p,
                                             activity_p)
    assert SX_STATUS_SUCCESS == rc, "Failed to get activity of MC route on VRID %d, rc: %d" % (vrid, rc)
    
    activity = boolean_t_p_value(activity_p)
    print "Activity of MC route is %s" % ("True" if (activity != 0) else "False")
    
def add_egress_rifs(vrid, mc_route_key, new_erifs):
    mc_route_key_p = new_sx_mc_route_key_t_p()
    sx_mc_route_key_t_p_assign(mc_route_key_p, mc_route_key)
    
    new_erifs_arr = make_egress_rif_list(new_erifs)
    
    rc = sx_api_router_mc_egress_rif_set(sx_api_router.handle,
                                         SX_ACCESS_CMD_ADD,
                                         vrid,
                                         mc_route_key_p,
                                         new_erifs_arr,
                                         len(new_erifs))
    assert SX_STATUS_SUCCESS == rc, "Failed to add %d egress RIFs to MC route on VRID %d, rc: %d" % (len(new_erifs), vrid, rc)
    print "Added %d egress RIFs to MC route on VRID %d" % (len(new_erifs), vrid)
    
def set_egress_rifs(vrid, mc_route_key, new_erifs):
    mc_route_key_p = new_sx_mc_route_key_t_p()
    sx_mc_route_key_t_p_assign(mc_route_key_p, mc_route_key)
    
    new_erifs_arr = make_egress_rif_list(new_erifs)
    
    rc = sx_api_router_mc_egress_rif_set(sx_api_router.handle,
                                         SX_ACCESS_CMD_SET,
                                         vrid,
                                         mc_route_key_p,
                                         new_erifs_arr,
                                         len(new_erifs))
    assert SX_STATUS_SUCCESS == rc, "Failed to set %d egress RIFs to MC route on VRID %d, rc: %d" % (len(new_erifs), vrid, rc)
    print "Set %d egress RIFs to MC route on VRID %d" % (len(new_erifs), vrid)
    
def delete_egress_rifs(vrid, mc_route_key, deleted_erifs):
    mc_route_key_p = new_sx_mc_route_key_t_p()
    sx_mc_route_key_t_p_assign(mc_route_key_p, mc_route_key)
    
    deleted_erifs_arr = make_egress_rif_list(deleted_erifs)
    
    rc = sx_api_router_mc_egress_rif_set(sx_api_router.handle,
                                         SX_ACCESS_CMD_DELETE,
                                         vrid,
                                         mc_route_key_p,
                                         deleted_erifs_arr,
                                         len(deleted_erifs))
    assert SX_STATUS_SUCCESS == rc, "Failed to delete %d egress RIFs from MC route on VRID %d, rc: %d" % (len(deleted_erifs), vrid, rc)
    print "Deleted %d egress RIFs from MC route on VRID %d" % (len(deleted_erifs), vrid)
    
def delete_all_egress_rifs(vrid, mc_route_key):
    mc_route_key_p = new_sx_mc_route_key_t_p()
    sx_mc_route_key_t_p_assign(mc_route_key_p, mc_route_key)
        
    rc = sx_api_router_mc_egress_rif_set(sx_api_router.handle,
                                         SX_ACCESS_CMD_DELETE_ALL,
                                         vrid,
                                         mc_route_key_p,
                                         None,
                                         0)
    assert SX_STATUS_SUCCESS == rc, "Failed to delete all egress RIFs from MC route on VRID %d, rc: %d" % (vrid, rc)
    print "Deleted all egress RIFs from MC route on VRID %d" % (vrid)    

def bind_mc_route_flow_counter(vrid, mc_route_key, counter_id):
    mc_route_key_p = new_sx_mc_route_key_t_p()
    sx_mc_route_key_t_p_assign(mc_route_key_p, mc_route_key)
        
    rc = sx_api_router_mc_route_counter_bind_set(sx_api_router.handle,
                                                 SX_ACCESS_CMD_BIND,
                                                 vrid,
                                                 mc_route_key_p,
                                                 counter_id)
    
    assert SX_STATUS_SUCCESS == rc, "Failed to bind MC route to counter %d on VRID %d, rc: %d" % (counter_id, vrid, rc)
    print "Binding MC route to counter %d on VRID %d" % (counter_id, vrid)        
    
def create_flow_counter():
    " This function creates ecmp flow counter. "

    counter_p = new_sx_flow_counter_id_t_p()
    counter_type = SX_FLOW_COUNTER_TYPE_PACKETS_AND_BYTES

    rc = sx_api_flow_counter_set(sx_api_router.handle, SX_ACCESS_CMD_CREATE, counter_type, counter_p)
    assert SX_STATUS_SUCCESS == rc, "Failed to create flow counter, rc: %d" % (rc)

    counter_id = sx_router_counter_id_t_p_value(counter_p)

    print "Created Flow counter %d" % (counter_id)
    return counter_id

def read_flow_counters(flow_counter_id):
     " This function reads flow counter. "
 
     counter_set = sx_flow_counter_set_t()
     counter_set_p = new_sx_flow_counter_set_t_p()
 
     rc = sx_api_flow_counter_get(sx_api_router.handle, SX_ACCESS_CMD_READ, flow_counter_id, counter_set_p)
     assert SX_STATUS_SUCCESS == rc, "Failed to read flow counter %d, rc: %d" % (flow_counter_id, rc) 
     print "Read flow counter %d" % (flow_counter_id)
     counter_set = sx_flow_counter_set_t_p_value(counter_set_p)
     return counter_set
 
def unbind_mc_route_flow_counter(vrid, mc_route_key):
    mc_route_key_p = new_sx_mc_route_key_t_p()
    sx_mc_route_key_t_p_assign(mc_route_key_p, mc_route_key)
        
    rc = sx_api_router_mc_route_counter_bind_set(sx_api_router.handle,
                                                 SX_ACCESS_CMD_UNBIND,
                                                 vrid,
                                                 mc_route_key_p,
                                                 0)
    
    assert SX_STATUS_SUCCESS == rc, "Failed to unbind flow counter from MC route VRID %d, rc: %d" % (vrid, rc)
    print "Unbinding flow counter from MC route on VRID %d" % (vrid)   

def destroy_flow_counter(flow_counter_id):
    " This function destroys flow counter. "

    counter_type = SX_FLOW_COUNTER_TYPE_PACKETS
    counter_p = new_sx_flow_counter_id_t_p()
    sx_flow_counter_id_t_p_assign(counter_p, flow_counter_id)

    rc = sx_api_flow_counter_set(sx_api_router.handle, SX_ACCESS_CMD_DESTROY, counter_type, counter_p)
    assert SX_STATUS_SUCCESS == rc, "Failed to destroy flow counter, rc: %d" % (rc)

    print "Destroyed flow counter %d" % (flow_counter_id)
        
def main():
    # Initialize router
    sx_api_router.router_init(ipv4_mc_enable = SX_ROUTER_ENABLE_STATE_ENABLE)
    
    # Create VRID that supports IPv4, both UC and MC    
    vrid = sx_api_router.create_vrid(ipv4_mc_enable = SX_ROUTER_ENABLE_STATE_ENABLE)
    
    port_vlan_list = [(PORT2, 3)]
    
    rif_arr = [None] * 4
    
    # Create 4 RIFs, one of each type that supports MC routing
    rif_arr[0] = prepare_vlan_rif(vrid, PORT1, 2, ether_addr(0x00, 0x02, 0x03, 0x04, 0x05, 0x07))
    vport_arr, bridge_id, rif_arr[1] = prepare_bridge_rif(vrid, port_vlan_list, ether_addr(0x00, 0x02, 0x03, 0x04, 0x05, 0x08))
    rif_arr[2] = prepare_router_port_rif(vrid, PORT3, ether_addr(0x00, 0x02, 0x03, 0x04, 0x05, 0x09))
    vport, rif_arr[3] = prepare_vport_rif(vrid, PORT4, 5, ether_addr(0x00, 0x02, 0x03, 0x04, 0x05, 0x0a))
     
    #Create flow counter
    counter_id = create_flow_counter()    
    
    # Enable IPv4 UC and MC on each RIF
    for i in xrange(4):
        sx_api_router.set_rif_state_ipv4(rif_arr[i], 
                                         ipv4_mc_enable = SX_ROUTER_ENABLE_STATE_ENABLE)
    
    erif_list = [rif_arr[2]]

    # Create MC route (S = 0.0.0.0/0, G = 224.1.1.1/32) with one egress RIF
    mc_route_key = make_ipv4_mc_route_key(DONT_CARE_RIF, '0.0.0.0', '0.0.0.0', '224.1.1.1', '255.255.255.255')
    mc_route_attr = make_mc_route_attributes()
    mc_route_data = make_mc_route_data(erif_list)
    add_mc_route(vrid, mc_route_key, mc_route_attr, mc_route_data)
    
    # Bind flow counter to mc route
    bind_mc_route_flow_counter(vrid, mc_route_key, counter_id)
    
    # Send traffic to 224.1.1.1
    
    # Read flow counters
    counter_set = read_flow_counters(counter_id)

    # Unbind flow counter from mc route
    unbind_mc_route_flow_counter(vrid, mc_route_key)
        
    # Create MC route (S = 192.168.1.2/32, G = 224.1.1.2/32) with one egress RIF
    mc_route_key = make_ipv4_mc_route_key(DONT_CARE_RIF, '192.168.1.2', '255.255.255.255', '224.1.1.2', '255.255.255.255')
    add_mc_route(vrid, mc_route_key, mc_route_attr, mc_route_data)
    
    # Send traffic from 192.168.1.2 to 224.1.1.2
    
    # Create directional MC route (S = 192.168.1.2/32, G = 224.1.1.3/32) with one egress RIF
    mc_route_key = make_ipv4_mc_route_key(rif_arr[0], '192.168.1.2', '255.255.255.255', '224.1.1.3', '255.255.255.255')
    mc_route_attr = make_mc_route_attributes(SX_ROUTER_RPF_ACTION_DIRECTIONAL)
    add_mc_route(vrid, mc_route_key, mc_route_attr, mc_route_data)
    
    # Send traffic from first RIF, from 192.168.1.2 to 224.1.1.3 

    # Add two additional egress RIFs to last MC route created    
    erif_list = [rif_arr[1], rif_arr[3]]
    add_egress_rifs(vrid, mc_route_key, erif_list)
    
    # Send traffic from first RIF, from 192.168.1.2 to 224.1.1.3 
    
    # Create MC route with RPF action DROP (S = 192.168.1.2/32, G = 224.1.1.4/32) with one egress RIF
    mc_route_key = make_ipv4_mc_route_key(rif_arr[0], '192.168.1.2', '255.255.255.255', '224.1.1.4', '255.255.255.255')
    mc_route_attr = make_mc_route_attributes(SX_ROUTER_RPF_ACTION_DROP)
    add_mc_route(vrid, mc_route_key, mc_route_attr, mc_route_data)
    
    # Send traffic from first RIF, from 192.168.1.2 to 224.1.1.4

    # Get activity of last MC route configured, without clearing activity (activity should be True)    
    get_mc_route_activity(vrid, mc_route_key, False)
    
    # Destroy flow counter
    destroy_flow_counter(counter_id)
    
    # Delete all MC routes
    delete_all_mc_routes(vrid)
    
    # Destroy all RIFs
    destroy_vport_rif(vrid, PORT4, 5, vport, rif_arr[3])
    destroy_router_port_rif(vrid, rif_arr[2])
    destroy_bridge_rif(vrid, vport_arr, bridge_id, rif_arr[1], port_vlan_list)
    destroy_vlan_rif(vrid, PORT1, 2, rif_arr[0])

    # Destroy VRID    
    sx_api_router.delete_vrid(vrid)
    
    # Deinitialize router
    sx_api_router.router_deinit()
 
if __name__ == "__main__":
    main()
