#!/usr/bin/env python
"""
This file contains Python command example for the Telemetry module.
Python commands syntax is very similar to the Switch SDK APIs. 
You can learn more about each command and its parameters by reading the SwitchX API Reference Manual.
"""

import sys, socket, struct, errno

from python_sdk_api.sx_api import *
import sx_api_router as sx_api_router

######################################################
#    functions 
######################################################

def tele_init():
    " This function init tele. "
    
    tele_param_p = new_sx_tele_init_params_t_p()

    rc = sx_api_tele_init_set(sx_api_router.handle, tele_param_p)
    assert SX_STATUS_SUCCESS == rc, "tele init failed rc: %d" % (rc) 
    print "sx_api_tele_init_set rc: %d " % (rc)    

def tele_deinit():
    " This function deinit tele. "

    rc = sx_api_tele_deinit_set(sx_api_router.handle)
    assert SX_STATUS_SUCCESS == rc, "tele deinit failed rc %d" % (rc) 
    print "sx_api_tele_deinit_set rc: %d " % (rc)  

def tele_histogram_set(cmd, port, tc, type, sample_time=0, min_boundary=0, resolution=0):
    " This function set histogram entry. "
    
    key = sx_tele_histogram_key_t()
    key.type = type
    
    que_key = sx_tele_histogram_queue_depth_key_t()
    que_key.network_port = port
    que_key.traffic_class = tc
    key.key.queue_depth = que_key
    
    attr = sx_tele_histogram_attributes_data_t()
    attr.type = type
    
    que_data = sx_tele_histogram_queue_depth_attributes_data_t()
    que_data.sample_time_resolution = sample_time
    que_data.min_boundary = min_boundary
    que_data.bin_size_resolution = resolution
    
    attr.data.queue_depth = que_data
    
    rc = sx_api_tele_histogram_set(sx_api_router.handle, cmd, key, attr)
    assert SX_STATUS_SUCCESS == rc, "sx_api_tele_histogram_set rc %d" % (rc) 
    print "sx_api_tele_histogram_set rc: %d " % (rc)

def tele_threshold_set(cmd, port, tc, threshold):
    " This function sets a threshold on the port and enables/disables the TC. "
    
    key = sx_tele_threshold_key_t()
    key.key_type = SX_TELE_THRESHOLD_TYPE_PORT_TC_E
    
    key_attr = sx_tele_threshold_port_tc_key_t()
    key_attr.log_port = port
    key_attr.tc = tc
    key.key.port_tc = key_attr
    
    data = sx_tele_threshold_data_t()
    data.data_type = SX_TELE_THRESHOLD_TYPE_PORT_TC_E
    data.data.threshold = threshold
    
    rc = sx_api_tele_threshold_set(sx_api_router.handle, cmd, key, data)
    assert SX_STATUS_SUCCESS == rc, "sx_api_tele_threshold_set rc %d" % (rc) 
    print "sx_api_tele_threshold_set rc: %d " % (rc)
    
def tele_threshold_get(cmd, port, tc):
    " This function gets the threshold of the port. "
    
    key = sx_tele_threshold_key_t()
    key.key_type = SX_TELE_THRESHOLD_TYPE_PORT_TC_E
    key.key.port_tc.log_port = port
    key.key.port_tc.tc = tc
    
    data_p = new_sx_tele_threshold_data_t_p()
    
    rc = sx_api_tele_threshold_get(sx_api_router.handle, SX_ACCESS_CMD_GET, key, data_p)
    assert SX_STATUS_SUCCESS == rc, "sx_api_tele_threshold_get rc %d" % (rc) 
    print "sx_api_tele_threshold_get rc: %d " % (rc)
        
    data = sx_tele_threshold_data_t_p_value(data_p)
    
    return data

def tele_threshold_crossed_data_get(cmd, port, tc):
    " This function gets the threshold crossing status of select keys. (above/below) - disabled TCs will always return \"below\" "
    
    crossed_data_p = new_sx_tele_threshold_crossed_data_t_p()
    key = sx_tele_threshold_key_t()
    key.key_type = SX_TELE_THRESHOLD_TYPE_PORT_TC_E
    
    key_attr = sx_tele_threshold_port_tc_key_t()
    key_attr.log_port = port
    key_attr.tc = tc
    key.key.port_tc = key_attr
    
    key_p = new_sx_tele_threshold_key_t_p()    
    sx_tele_threshold_key_t_p_assign(key_p, key)
    
    rc = sx_api_tele_threshold_crossed_data_get(sx_api_router.handle, cmd, key, crossed_data_p, 1)
    assert SX_STATUS_SUCCESS == rc, "sx_api_tele_threshold_crossed_data_get rc %d" % (rc) 
    print "sx_api_tele_threshold_crossed_data_get rc: %d " % (rc)
    
    return sx_tele_threshold_crossed_data_t_p_value(crossed_data_p).data.port_tc

def tele_histogram_get(port, tc, type):
    " This function get histogram entry. "
    
    key = sx_tele_histogram_key_t()
    key.type = type
    
    que_key = sx_tele_histogram_queue_depth_key_t()
    que_key.network_port = port
    que_key.traffic_class = tc
    key.key.queue_depth = que_key
    
    attr_p = new_sx_tele_histogram_attributes_data_t_p()
    
    rc = sx_api_tele_histogram_get(sx_api_router.handle, SX_ACCESS_CMD_GET, key, attr_p)
    assert SX_STATUS_SUCCESS == rc, "sx_api_tele_histogram_get rc %d" % (rc) 
    print "sx_api_tele_histogram_get rc: %d " % (rc)
        
    attr = sx_tele_histogram_attributes_data_t_p_value(attr_p)
    
    return attr

def tele_histogram_iter_get(cmd, hist_count, port=0, tc=0, type=0 , filter_port_valid=False, filter_port=0,
                            filter_tc_valid=False, filter_tc=0):
    " This function iter get histogram entry. "

    hist_list = new_sx_tele_histogram_key_t_arr(hist_count)   
              
    hist_count_p = new_uint32_t_p()    
    uint32_t_p_assign(hist_count_p, hist_count)    
     
    filter = sx_tele_histogram_filter_t()    
    if filter_port_valid==True:
        filter.filter_port_valid = SX_TELE_HISTOGRAM_KEY_FILTER_FIELD_VALID_E
        filter.port_filter = filter_port
    else:
        filter.filter_port_valid = SX_TELE_HISTOGRAM_KEY_FILTER_FIELD_NOT_VALID_E

    if filter_tc_valid==True:
        filter.filter_tc_valid = SX_TELE_HISTOGRAM_KEY_FILTER_FIELD_VALID_E
        filter.tc_filter = filter_tc
    else:
        filter.filter_tc_valid = SX_TELE_HISTOGRAM_KEY_FILTER_FIELD_NOT_VALID_E
                
    if port==0:
        key_p = None
    else:       
        key = sx_tele_histogram_key_t()
        key.type = type
    
        que_key = sx_tele_histogram_queue_depth_key_t()
        que_key.network_port = port
        que_key.traffic_class = tc
        key.key.queue_depth = que_key
        
        key_p = new_sx_tele_histogram_key_t_p()    
        sx_tele_histogram_key_t_p_assign(key_p, key)    
        
    rc = sx_api_tele_histogram_iter_get(sx_api_router.handle, cmd, key_p, filter, hist_list, hist_count_p)
    assert SX_STATUS_SUCCESS == rc, "sx_api_tele_histogram_iter_get rc %d" % (rc) 
    print "sx_api_tele_histogram_iter_get rc: %d " % (rc)
    
    count = uint32_t_p_value(hist_count_p)
    
    if hist_count > 0:    
        print "\n##### Histogram list #####\n"    
        for index in range (count):
            key = sx_tele_histogram_key_t_arr_getitem(hist_list, index) 
            que_key = key.key.queue_depth
            print "Key %d: Port 0x%x, TC %d" % (index, que_key.network_port, que_key.traffic_class)
        print "\n" 
           
    return count

def tele_histogram_data_get(port, tc, type, clear=False):
    " This function get histogram entry. "
    
    hist_data = sx_tele_histogram_data_t()
    hist_data_p = new_sx_tele_histogram_data_t_p()   
     
    bins = new_uint64_t_arr(SX_TELE_HIST_MAX_BINS)
    
    hist_data.bins = bins
    hist_data.bins_cnt = SX_TELE_HIST_MAX_BINS
    sx_tele_histogram_data_t_p_assign(hist_data_p, hist_data) 
    
    key = sx_tele_histogram_key_t()
    key.type = type
    
    que_key = sx_tele_histogram_queue_depth_key_t()
    que_key.network_port = port
    que_key.traffic_class = tc
    key.key.queue_depth = que_key

    if clear == True:
        cmd = SX_ACCESS_CMD_READ_CLEAR
    else:
        cmd = SX_ACCESS_CMD_READ

    rc = sx_api_tele_histogram_data_get(sx_api_router.handle, cmd, key, hist_data_p)
    assert SX_STATUS_SUCCESS == rc, "sx_api_tele_histogram_data_get rc %d" % (rc) 
    print "sx_api_tele_histogram_data_get rc: %d " % (rc)
    
    hist_data = sx_tele_histogram_data_t_p_value(hist_data_p) 
    bins =  hist_data.bins
    
    print "\n##### Histogram data #####\n"    
    for index in range (SX_TELE_HIST_MAX_BINS):
        val = uint64_t_arr_getitem(bins, index) 
        print " %d: %d" % (index, val)
    print "\n" 
 
#tele flow
def tele_threshold_flow(vrid, rif_arr):
    
    #add local and ip2me routes
    sx_api_router.create_local_route(vrid, rif_arr[0], "192.168.1.0", "255.255.255.0")
    sx_api_router.create_ip2me_route(vrid, rif_arr[0], "192.168.1.1")
    sx_api_router.create_local_route(vrid, rif_arr[1], "22.33.0.0", "255.255.0.0")
    sx_api_router.create_ip2me_route(vrid, rif_arr[1], "22.33.44.55")

    #rif state ipv4
    sx_api_router.set_rif_state_ipv4(rif_arr[0])
    sx_api_router.set_rif_state_ipv4(rif_arr[1])

    #add neighbors
    sx_api_router.add_neigh(rif_arr[1], "22.33.0.2", sx_api_router.NEIGH_MAC_2)
    
    #add uc route next_hop
    sx_api_router.add_uc_route_next_hop(vrid, "100.100.128.0", "255.255.255.0", {"22.33.0.2"})

    ####################
    ## Threshold flow ##
    ####################
    
    #init telemetry
    tele_init()
       
    #add threshold entry
    tele_threshold_set(cmd = SX_ACCESS_CMD_SET,
                       port = sx_api_router.PORT1, 
                       tc = 1,
                       threshold = 100)   
    
    #edit threshold entry
    tele_threshold_set(cmd = SX_ACCESS_CMD_EDIT,
                       port = sx_api_router.PORT1, 
                       tc = 1,
                       threshold = 150)    

    #add threshold entry
    tele_threshold_set(cmd = SX_ACCESS_CMD_SET,
                       port = sx_api_router.PORT2, 
                       tc = 4,
                       threshold = 200)    

    #get threshold data
    tele_attr = tele_threshold_get(cmd = SX_ACCESS_CMD_GET,
                                   port=sx_api_router.PORT1,
                                   tc=1)

    ##########################################################
    ## Traffic 192.168.1.2 --> 100.100.128.X(via 22.33.0.2)
    ##########################################################

    #read threshold data
    tele_threshold_get(cmd = SX_ACCESS_CMD_GET,
                       port = sx_api_router.PORT1, 
                       tc = 1)
        
    #delete threshold entry
    tele_threshold_set(cmd = SX_ACCESS_CMD_DESTROY,
                       port = sx_api_router.PORT1, 
                       tc = 1,
                       threshold = 100)    

    tele_threshold_set(cmd = SX_ACCESS_CMD_DESTROY,
                       port = sx_api_router.PORT2, 
                       tc = 4,
                       threshold = 150)    

    #deinit telemetry
    tele_deinit()
            
    #delete all next hop uc routes
    sx_api_router.delete_all_next_hops_uc_routes_per_vrid(vrid)

    #delete all neighbors per rif
    sx_api_router.delete_all_neigh_per_rif(rif_arr[0])
    sx_api_router.delete_all_neigh_per_rif(rif_arr[1])

    #delete all local and ip2me routes
    sx_api_router.delete_all_local_routes(vrid)
    sx_api_router.delete_all_ip2me_routes(vrid)

    #delete rif
    sx_api_router.delete_rif(vrid, rif_arr[0])
    sx_api_router.delete_rif(vrid, rif_arr[1])

def tele_flow(vrid, rif_arr):
    
    #add local and ip2me routes
    sx_api_router.create_local_route(vrid, rif_arr[0], "192.168.1.0", "255.255.255.0")
    sx_api_router.create_ip2me_route(vrid, rif_arr[0], "192.168.1.1")
    sx_api_router.create_local_route(vrid, rif_arr[1], "22.33.0.0", "255.255.0.0")
    sx_api_router.create_ip2me_route(vrid, rif_arr[1], "22.33.44.55")

    #rif state ipv4
    sx_api_router.set_rif_state_ipv4(rif_arr[0])
    sx_api_router.set_rif_state_ipv4(rif_arr[1])

    #add neighbors
    sx_api_router.add_neigh(rif_arr[1], "22.33.0.2", sx_api_router.NEIGH_MAC_2)
    
    #add uc route next_hop
    sx_api_router.add_uc_route_next_hop(vrid, "100.100.128.0", "255.255.255.0", {"22.33.0.2"})

    ####################
    ## Histogram flow ##
    ####################

    #init telemetry
    tele_init()
       
    #add histogram entry
    tele_histogram_set(cmd = SX_ACCESS_CMD_SET,
                       port = sx_api_router.PORT1, 
                       tc = 1,
                       type = SX_TELE_HISTOGRAM_TYPE_QUEUE_DEPTH_E,
                       sample_time = 3, 
                       min_boundary = 3,
                       resolution = 3)   
    
    #edit histogram entry
    tele_histogram_set(cmd = SX_ACCESS_CMD_EDIT,
                       port = sx_api_router.PORT1, 
                       tc = 1,
                       type = SX_TELE_HISTOGRAM_TYPE_QUEUE_DEPTH_E,
                       sample_time = 4, 
                       min_boundary = 4,
                       resolution = 4)    

    #add histogram entry
    tele_histogram_set(cmd = SX_ACCESS_CMD_SET,
                       port = sx_api_router.PORT2, 
                       tc = 4,
                       type = SX_TELE_HISTOGRAM_TYPE_QUEUE_DEPTH_E,
                       sample_time = 3, 
                       min_boundary = 3,
                       resolution = 3)    

    #get histogram attributes
    tele_attr = tele_histogram_get(port = sx_api_router.PORT1, 
                                   tc = 1,
                                   type = SX_TELE_HISTOGRAM_TYPE_QUEUE_DEPTH_E)

    #get histograms total size
    count = tele_histogram_iter_get(cmd = SX_ACCESS_CMD_GET, hist_count = 0) 

    #get histogram key list
    count = tele_histogram_iter_get(cmd = SX_ACCESS_CMD_GET_FIRST, hist_count = 2) 
    
    ##########################################################
    ## Traffic 192.168.1.2 --> 100.100.128.X(via 22.33.0.2)
    ##########################################################

    #read histogram data
    tele_histogram_data_get(port = sx_api_router.PORT1, 
                            tc = 1,
                            type = SX_TELE_HISTOGRAM_TYPE_QUEUE_DEPTH_E)
        
    #delete histogram entry
    tele_histogram_set(cmd = SX_ACCESS_CMD_DESTROY,
                       port = sx_api_router.PORT1, 
                       tc = 1,
                       type = SX_TELE_HISTOGRAM_TYPE_QUEUE_DEPTH_E)    

    tele_histogram_set(cmd = SX_ACCESS_CMD_DESTROY,
                       port = sx_api_router.PORT2, 
                       tc = 4,
                       type = SX_TELE_HISTOGRAM_TYPE_QUEUE_DEPTH_E)    



    ####################
    ## Threshold flow ##
    ####################
       
    #add threshold entry
    tele_threshold_set(cmd = SX_ACCESS_CMD_SET,
                       port = sx_api_router.PORT1, 
                       tc = 1,
                       threshold = 100)   
    
    #edit threshold entry
    tele_threshold_set(cmd = SX_ACCESS_CMD_EDIT,
                       port = sx_api_router.PORT1, 
                       tc = 1,
                       threshold = 1)    

    #add threshold entry
    tele_threshold_set(cmd = SX_ACCESS_CMD_SET,
                       port = sx_api_router.PORT2, 
                       tc = 4,
                       threshold = 200)    

    #get threshold data
    tele_attr = tele_threshold_get(cmd = SX_ACCESS_CMD_GET,
                                   port=sx_api_router.PORT1,
                                   tc=1)

    crossed_data = tele_threshold_crossed_data_get(cmd = SX_ACCESS_CMD_GET,
                                                   port=sx_api_router.PORT1,
                                                   tc=1)
    
    ##########################################################
    ## Traffic 192.168.1.2 --> 100.100.128.X(via 22.33.0.2)
    ##########################################################

    crossed_data = tele_threshold_crossed_data_get(cmd = SX_ACCESS_CMD_GET,
                                                   port=sx_api_router.PORT1,
                                                   tc=1)

    #read threshold data
    tele_threshold_get(cmd = SX_ACCESS_CMD_GET,
                       port = sx_api_router.PORT1, 
                       tc = 1)
        
    #delete threshold entry
    tele_threshold_set(cmd = SX_ACCESS_CMD_DESTROY,
                       port = sx_api_router.PORT1, 
                       tc = 1,
                       threshold = 100)    

    tele_threshold_set(cmd = SX_ACCESS_CMD_DESTROY,
                       port = sx_api_router.PORT2, 
                       tc = 4,
                       threshold = 150)
    
    
    #deinit telemetry
    tele_deinit()
            
    #delete all next hop uc routes
    sx_api_router.delete_all_next_hops_uc_routes_per_vrid(vrid)

    #delete all neighbors per rif
    sx_api_router.delete_all_neigh_per_rif(rif_arr[0])
    sx_api_router.delete_all_neigh_per_rif(rif_arr[1])

    #delete all local and ip2me routes
    sx_api_router.delete_all_local_routes(vrid)
    sx_api_router.delete_all_ip2me_routes(vrid)

    #delete rif
    sx_api_router.delete_rif(vrid, rif_arr[0])
    sx_api_router.delete_rif(vrid, rif_arr[1])

######################################################
#    main
######################################################
def main():

    #init router
    sx_api_router.router_init()
    
    vrid = sx_api_router.create_vrid()
    
    #vlan rif example
    rif_arr = sx_api_router.example_vlan_rif_init(vrid)
    
    tele_flow(vrid, rif_arr)
    
    sx_api_router.example_vlan_rif_deinit(vrid, rif_arr)
    
    #delete vrid
    sx_api_router.delete_vrid(vrid)
    
    #deinit router
    sx_api_router.router_deinit()

if __name__ == "__main__":
    main()
